import numpy as np
import pandas as pd
from datetime import datetime
import requests
from io import BytesIO
import zipfile
import os
from .makeBivSortInd import makeBivSortInd
from .runBivSort import runBivSort
import statsmodels.api as sm

def makeCOMPUSTATDerivedVariables(params):
    # Timekeeping
    print(f"\nNow working on making variables derived from COMPUSTAT. Run started at {datetime.now()}.\n")

    # Set dataPaths
    crsp_path = params.crspFolder + os.sep
    comp_path = params.compFolder + os.sep

    # Make Operating costs
    cogs = pd.read_csv(comp_path + 'COGS.csv', index_col=0)
    xsga = pd.read_csv(comp_path + 'XSGA.csv', index_col=0)
    OC = cogs + xsga
    OC.to_csv(comp_path + 'OC.csv')

    # Make Total Debt
    dlc = pd.read_csv(comp_path + 'DLC.csv', index_col=0)
    dltt = pd.read_csv(comp_path + 'DLTT.csv', index_col=0)
    D = dlc + dltt
    D.to_csv(comp_path + 'D.csv')

    # Make Investment. (Cumulative) Investment = property plant and equipment
    ppegt = pd.read_csv(comp_path + 'PPEGT.csv', index_col=0)
    invt = pd.read_csv(comp_path + 'INVT.csv', index_col=0)
    INV = ppegt + invt
    INV.to_csv(comp_path + 'INV.csv')

    # Make free cash-flow = net earnings + depreciation and amortization - changes in working capital - capex
    ni = pd.read_csv(comp_path +'NI.csv', index_col=0)
    dp = pd.read_csv(comp_path + 'DP.csv', index_col=0)
    wcapch = pd.read_csv(comp_path + 'WCAPCH.csv', index_col=0)
    capx = pd.read_csv(comp_path + 'CAPX.csv', index_col=0)
    FCF = ni + dp - wcapch - capx
    FCF.to_csv(comp_path + 'FCF.csv')

    # Make dividends (really it's money returned to equity holders (dividends and buy-backs)
    prstkc = pd.read_csv(comp_path + 'PRSTKC.csv', index_col=0)
    prstkc = prstkc.astype(float).fillna(0)
    prstkpc = pd.read_csv(comp_path + 'PRSTKPC.csv', index_col=0)
    prstkpc = prstkpc.astype(float).fillna(0)
    prstkcc = pd.read_csv(comp_path + 'PRSTKCC.csv', index_col=0)
    prstkcc = prstkcc.astype(float).fillna(0)
    dvc = pd.read_csv(comp_path + 'DVC.csv', index_col=0)
    dvc = dvc.astype(float).fillna(0)
    temp = prstkc - prstkpc
    prstkcc = prstkcc.combine_first(temp) # this combines PRSTKCC with the temp dataframe, such that any missing values in PRSTKCC are replaced by the corresponding values in 'temp'
    DIV = dvc + prstkcc
    DIV.to_csv(comp_path + 'DIV.csv')

    # Make Working capital
    act = pd.read_csv(comp_path + 'ACT.csv', index_col=0)
    lct = pd.read_csv(comp_path + 'LCT.csv', index_col=0)
    WCAP = act - lct
    WCAP.to_csv(comp_path + 'WCAP.csv')

    # Make CFO (cash flow from operations)
    OANCF = pd.read_csv(comp_path + 'OANCF.csv', index_col=0)
    IB = pd.read_csv(comp_path + 'IB.csv', index_col=0)
    DP = pd.read_csv(comp_path + 'DP.csv', index_col=0)
    XIDO = pd.read_csv(comp_path + 'XIDO.csv', index_col=0)
    TXDC = pd.read_csv(comp_path + 'TXDC.csv', index_col=0)
    ESUBC = pd.read_csv(comp_path + 'ESUBC.csv', index_col=0)
    SPPIV = pd.read_csv(comp_path + 'SPPIV.csv', index_col=0)
    FOPO = pd.read_csv(comp_path + 'FOPO.csv', index_col=0)
    RECCH = pd.read_csv(comp_path + 'RECCH.csv', index_col=0)
    INVCH = pd.read_csv(comp_path + 'INVCH.csv', index_col=0)
    APALCH = pd.read_csv(comp_path + 'APALCH.csv', index_col=0)
    TXACH = pd.read_csv(comp_path + 'TXACH.csv', index_col=0)
    AOLOCH = pd.read_csv(comp_path + 'AOLOCH.csv', index_col=0)
    temp = IB + DP # Income before extraordinary items + depreciation and amortization
    temp = temp + XIDO # + extraordinary items and discontinued operations
    temp = temp + TXDC # + deferred taxes (CF)
    temp = temp + ESUBC # + equity in net loss (Earnings)
    temp = temp - SPPIV # - sale of PPE and sale of investments gain (loss)
    temp = temp + FOPO # + funds from operations - Other
    temp = temp + RECCH # + accounts receivable - decrease (increase)
    temp = temp + INVCH # + inventory - decrease (increase)
    temp = temp + APALCH # + accounts payable and accrued liabilities - increase (decrease)
    temp = temp + TXACH # + income taxes - acctued - increase (decrease)
    temp = temp + AOLOCH # + assets and liabi.ities - other (net change)
    CFO = temp.copy()  # operating activities - net cash flow (i.e., CFFO)
    CFO = CFO.combine_first(OANCF)
    CFO.to_csv(comp_path + 'CFO.csv')

    # Make book equity. Start with shareholder's equity
    SEQ = pd.read_csv(comp_path + 'SEQ.csv', index_col=0)
    CEQ = pd.read_csv(comp_path + 'CEQ.csv', index_col=0)
    PSTK = pd.read_csv(comp_path + 'PSTK.csv', index_col=0)
    AT = pd.read_csv(comp_path + 'AT.csv', index_col=0)
    LT = pd.read_csv(comp_path + 'LT.csv', index_col=0)
    SE = SEQ.copy()                          # Shareholder equity
    temp = CEQ + PSTK
    SE[SE.isna()] = temp[SE.isna()]         # Uses common equity + preferred stock if SEQ is missing
    temp = AT - LT
    SE[SE.isna()] = temp[SE.isna()]         # Uses assets - liabilities, if others are missing

    # Make preferred stock
    PSTKRV = pd.read_csv(comp_path + 'PSTKRV.csv', index_col=0)
    PSTKL = pd.read_csv(comp_path + 'PSTKL.csv', index_col=0)
    PSTK = pd.read_csv(comp_path + 'PSTK.csv', index_col=0)
    PS = PSTKRV.copy()
    PS[PS.isna()] = PSTKL[PS.isna()]
    PS[PS.isna()] = PSTK[PS.isna()]
    PS.to_csv(comp_path + 'PS.csv')

    # Make Deferred taxes
    TXDITC = pd.read_csv(comp_path + 'TXDITC.csv', index_col=0)
    TXDB = pd.read_csv(comp_path + 'TXDB.csv', index_col=0)
    ITCB = pd.read_csv(comp_path + 'ITCB.csv', index_col=0)
    DT = TXDITC.copy()
    temp = TXDB + ITCB
    DT[DT.isna()] = temp[DT.isna()]

    # Book equity is shareholder equity + deferred taxes - preferred stock
    BE = SE + (DT - PS)

    "Add in the Davis, Fama and French (1997) book equities (from before the accounting data is available on COMPUSTAT)"
    # First load the dates and permno vectors
    # dates = pd.read_csv(comp_path + 'dates.csv', index_col=0)
    # permno = pd.read_csv(comp_path + 'permno.csv', index_col=0)
    ffurl = 'http://mba.tuck.dartmouth.edu/pages/faculty/ken.french/ftp/Historical_BE_Data.zip'

    # tricks the host site into thinking the HTTP call is from user instead of machine
    headers = {"User-Agent": "Mozilla/5.0"}
    response = requests.get(ffurl, headers=headers)
    # saving the zip file as an object
    zf = zipfile.ZipFile(BytesIO(response.content))
    # extracting and saving the contents
    zf.extractall(comp_path)
    # loading the txt file that was extracted.
    ffdata = pd.read_csv(comp_path + 'DFF_BE_With_Nonindust.txt', header=None)

    # Split the data into columns
    ffdata = ffdata[0].str.split(' ', expand=True)
    # replace all empty cells with NaN values; keep the -99.99s for now
    ffdata.replace('', np.nan, inplace=True)
    ffdata = ffdata.apply(pd.to_numeric, errors='coerce')

    # create a new dataframe to store rearranged values
    temp_df = pd.DataFrame()

    # loop through each row and collect only non NaN values.
    for row in ffdata.values:
        # get the non-NaN values
        new_row = row[~np.isnan(row)]

        # append the rearranged row to the new dataframe
        temp_df = temp_df.append(pd.Series(new_row), ignore_index=True)

    # cleaned data
    ffdata = temp_df.copy()

    # Rename the columns
    ffdata.columns = ['permno', 'start', 'end'] + list(range(1, ffdata.shape[1]-2))

    # Set permno as the index
    ffdata = ffdata.set_index('permno')

    # Convert start and end to integers
    ffdata['start'] = ffdata['start'].astype(int)
    ffdata['end'] = ffdata['end'].astype(int)

    # Create a new DataFrame with the years as columns
    newdata = pd.DataFrame(index=ffdata.index, columns=[str(year) for year in range(ffdata['start'].min(),
                                                                                    ffdata['end'].max()+1)])
    # newdata = pd.DataFrame(index=ffdata.index, columns=[str(year*100 + 6) for year in range(ffdata['start'].min(), ffdata['end'].max()+1)])

    # Fill in the new DataFrame with the values corresponding to the correct years
    for perm in ffdata.index:
        start = ffdata.loc[perm, 'start']
        end = ffdata.loc[perm, 'end']
        newdata.loc[perm, str(start):str(end)] = ffdata.loc[perm, list(range(1, end-start+2))].values

    # Convert the values to numeric
    newdata = newdata.apply(pd.to_numeric)

    # Replacing the -99.99 values with NaN
    newdata = newdata.replace(-99.99, np.nan)
    ffdata = newdata.copy()

    # Renaming the columns
    ffdata.columns = [i + '06.0' for i in ffdata.columns]
    ffdata.columns = ffdata.columns.astype(float)

    # Putting ffdata in date x permno form. Then change the column names to be the same datatype (str) as BE columns
    ffdata = ffdata.T
    ffdata.columns = ffdata.columns.astype(str)

    # Identify overlapping firms
    BE.columns = BE.columns.astype(int)  # Make sure both have the same column datatypes
    ffdata.columns = ffdata.columns.astype(float).astype(int)
    common_permno = BE.columns.intersection(ffdata.columns)

    # ffdata has a float index but BE has int index.
    ffdata.index = ffdata.index.astype(int)

    # Get the overlapping dates
    common_dates = BE.index.intersection(ffdata.index)

    # Update BE with data from ffdata for the common firms and additional dates
    for permno in common_permno:
        BE.loc[common_dates, permno] = ffdata.loc[common_dates, permno]

    # Save the extended book equity dataframe, which we rename back to BE
    BE.to_csv(comp_path + 'BE.csv')

    # load market value of equity and set the columns to be integers
    me = pd.read_csv(crsp_path + 'me.csv', index_col=0)
    me.columns = me.columns.values.astype(float).astype(int)

    # Following Fama-French 1993, we divide each value in BE by the value of me six months ago. This avoids a short
    # momentum position from sneaking into value strategies.
    # -- First I need to find the permnos in me that are also in BE since BE has a lot more permnos
    # common_permno = BE.columns.intersection(me.columns)

    # First make sure me and be have the same indices and coluns
    BE.columns = me.columns
    BE.index = me.index

    # Divide 'BE' by Shifted 'me' for Common Firms
    bm = BE.divide(me.shift(6), axis=0)

    # Save the book market dataframe
    bm.to_csv(comp_path + 'BM.csv')

    # The assets of financial firms are very different (and larger than) those of other firms; when we scale by assets we'll
    # often want to kick out financial
    SIC = pd.read_csv(crsp_path + 'siccd.csv', index_col=0)
    FinFirms = (SIC >= 6000) & (SIC <= 6999) # creates a boolean mask that is 1 when the SIC is in the 6000's.
    # uses the .any() method with axis=0 to check which columns have at least one True value (i.e., meet the condition) and
    FinFirms = FinFirms.columns[FinFirms.any(axis=0)].values
    # Changes FinFirms to a dataframe
    FinFirms = pd.DataFrame(FinFirms)
    # Set the column label
    FinFirms.columns = ['permno']
    # Save to path
    FinFirms.to_csv(comp_path + 'FinFirms.csv')

    # Compare our HML factor with that from Ken French's website
    NYSE = pd.read_csv(crsp_path + 'NYSE.csv', index_col=0).astype(int)
    ret = pd.read_csv(crsp_path + 'ret.csv', index_col=0).astype(float)
    bm = pd.read_csv(comp_path + 'BM.csv', index_col=0).astype(float)
    ff = pd.read_csv(crsp_path + 'ff.csv', index_col=0).astype(float)
    dates = pd.read_csv(crsp_path + 'dates.csv', index_col=0).astype(int)

    # Fama and French kick out the negative book-equity firms (the most growthy)
    bm[bm < 0] = np.nan
    ind = makeBivSortInd(me, 2, bm, [30, 70], breaksFilterInd=NYSE)
    res = runBivSort(params, ret, ind, 2, 3, me, dates, printResults=0)

    # Replicate HML from the corner portfolios
    hmlrep = (res[0]['pret'][:, 2] + res[0]['pret'][:, 5] - res[0]['pret'][:, 0] - res[0]['pret'][:, 3])/2

    # Should see a high (~99%) correlation, but it WON'T be perfect for a couple of reasons. Fama-French construct the
    # variable using data available at the time of construction (they don't go back and change it if the fix data errors
    # in crsp/comp). It shouldn't matter: "value" is a robust phenomena -- one of the reasons it is important is that
    # the details don't "matter" for getting a real "exposure" to value.

    # Let the user know of the check.
    print("\nNow let's compare our HML with that form Ken French's website.")

    # Correlation should be >95%
    hml = ff['hml'].to_numpy()
    index = np.isfinite(hmlrep + hml)
    correlation = np.corrcoef(hmlrep[index], hml[index])[0, 1]

    # Print the correlation
    print(f"The correlation between HML from Ken French and replicated HML is {correlation * 100:.2f}%.")

    # Calculate the mean return of HML and replicated HML
    mean_hml = np.nanmean(hml)
    mean_hmlrep = np.nanmean(hmlrep)
    print(f"\nMean return of HML: {mean_hml:.2f}% per month")
    print(f"Mean return of replicated HML: {mean_hmlrep:.2f}% per month")

    # Regress HML on replicated HML
    model1 = sm.OLS(hml, sm.add_constant(hmlrep), missing='drop').fit()
    print('\nRegress HML on replicated HML:')
    print(model1.summary())

    # Regress replicated HML on HML
    model2 = sm.OLS(hmlrep, sm.add_constant(hml), missing='drop').fit()
    print('\nRegress replicated HML on HML:')
    print(model2.summary())

    # Make quarterly book equity. Start with shareholder's equity
    SEQQ = pd.read_csv(comp_path + 'SEQQ.csv', index_col=0)  # Shareholder equity
    CEQQ = pd.read_csv(comp_path + 'CEQQ.csv', index_col=0)  # Common equity
    PSTKQ = pd.read_csv(comp_path + 'PSTKQ.csv', index_col=0)  # Preferred stock
    ATQ = pd.read_csv(comp_path + 'ATQ.csv', index_col=0)  # Assets
    LTQ = pd.read_csv(comp_path + 'LTQ.csv', index_col=0)  # Liabilities
    DT = pd.read_csv(comp_path + 'TXDITCQ.csv', index_col=0)  # Deferred taxes

    SE = SEQQ.copy()  # Initialize SE with SEQQ

    # Update SE where it's NaN with the sum of common equity (CEQQ) and preferred stock (PSTKQ)
    SE.update(CEQQ.add(PSTKQ, fill_value=0), overwrite=False)

    # Further update SE where it's still NaN with the difference between assets (ATQ) and liabilities (LTQ)
    SE.update(ATQ.subtract(LTQ, fill_value=0), overwrite=False)

    # Book equity is shareholder equity + deferred taxes - preferred stock
    BEQ = SE + DT - PSTKQ
    BEQ.to_csv(comp_path + 'BEQ.csv')

    # Make quarterly gross profitability
    REVTQ = pd.read_csv(comp_path + 'REVTQ.csv', index_col=0)
    COGSQ = pd.read_csv(comp_path + 'COGSQ.csv', index_col=0)
    GPQ = REVTQ - COGSQ
    GPQ.to_csv(comp_path + 'GPQ.csv')

    # Make quarterly return-on-assets
    IBQ = pd.read_csv(comp_path + 'IBQ.csv', index_col=0)
    roa = IBQ / ATQ.shift(3, axis=0)
    roa.to_csv(comp_path + 'roa.csv')

    # SUE and dROE
    EPSPXQ = pd.read_csv(comp_path + 'EPSPXQ.csv', index_col=0)

    # The commented out section below works but is very very slow.
    # Find the report dates for IBQ and create a boolean matrix
    # idxRprtDate = (IBQ != IBQ.shift(1)).fillna(False) & IBQ.notna()
    #
    # # Initialize variables for the quarterly lags
    # BEQL1 = pd.DataFrame(np.nan, index=IBQ.index, columns=IBQ.columns)
    # IBQL4 = BEQL1.copy()
    #
    # # Create IBQ and EPSPXQ only for the report dates
    # IBQL0 = IBQ.where(idxRprtDate)
    # EPSPXQL0 = EPSPXQ.where(idxRprtDate)
    #
    # # Initiate the index of the number of lagged report dates available
    # numLagIdx = idxRprtDate.astype(int)
    #
    # # Dictionary to store lagged EPSPXQ data
    # epsStruct = {i: pd.DataFrame(np.nan, index=IBQ.index, columns=IBQ.columns) for i in range(1, 13)}
    #
    # # Loop over the past 60 months
    # for i in range(1, 61):
    #     numLagIdx += idxRprtDate.shift(i).fillna(False).astype(int)
    #     thisNumLagIdx = numLagIdx * idxRprtDate.shift(i).fillna(False).astype(int)
    #
    #     lIBQ = IBQ.shift(i)
    #     lBEQ = BEQ.shift(i)
    #     lEPSPXQ = EPSPXQ.shift(i)
    #
    #     IBQL4[thisNumLagIdx == 5] = lIBQ[thisNumLagIdx == 5]
    #     BEQL1[thisNumLagIdx == 2] = lBEQ[thisNumLagIdx == 2]
    #
    #     for j in range(1, 13):
    #         epsStruct[j][thisNumLagIdx == j + 1] = lEPSPXQ[thisNumLagIdx == j + 1]

    # Timekeeping

    # Create lagged data for different quarters

    # Function to fill in months between announcements

    # Fill in the months in between announcements

    start_date = IBQ.index.get_loc(197101.0, method='bfill')

    # Initialize SUE dataframe with NaN
    SUE = pd.DataFrame(index=IBQ.index, columns=IBQ.columns).astype(float)

    # Difference between quarterly EPS and the EPS of the same quarter last year
    EPS_diff = EPSPXQ - EPSPXQ.shift(12, axis=0)

    # Rolling 2 year standard deviations; requiring at least 6 observations
    EPS_std = EPS_diff.rolling(window=24, min_periods=6).std()

    # Dividing the eps difference by the 2 year standard deviations
    SUE = EPS_diff / EPS_std

    # Forward fill missing data and save it
    SUE = SUE.ffill()
    SUE.to_csv(comp_path + 'SUE.csv')

    # Create and store dROE, which equals the annual change in net income divided by the previous quarters shareholder's
    # equity
    dROE = (IBQ - IBQ.shift(12)) / BEQ.shift(4)
    dROE = dROE.ffill()
    dROE.to_csv(comp_path + 'dROE.csv')

    # Make SUE2 -- a simpler construction
    # current earnings minus earnings from one year ago, scaled by the stand. dev. of last 8 quarterly earnings.
    SUE2 = (IBQ - IBQ.shift(12)) / IBQ.rolling(window=24, min_periods=2).std()
    SUE2 = SUE2.ffill()
    SUE2.to_csv(comp_path + 'SUE2.csv')

    # Make dROA -- difference between current earnings and average of the last four quarters, scaled by lagged assets
    # :TODO Ask Mish if ATQ should be lagged or not. The description says to but it doesn't appear in Matlab code.
    # :TODO Also ask Mish if we should forward fill dROA like we did for dROE.
    dROA = (IBQ - IBQ.rolling(window=12, min_periods=2).mean()) / ATQ
    dROA = dROA.ffill()
    dROA.to_csv(comp_path + 'dROA.csv')

    # announcement month effect (Frazzini-Lamont)
    amonth = IBQ != IBQ.shift(1)

    # Filter to ensure that changes are only flagged where IBQ has valid data
    amonth = amonth & IBQ.notna()
    amonth.to_csv(comp_path + 'amonth.csv')

    # from AssayingAnomalies.Functions.runUnivSort import runUnivSort
    # indAMO = IBQ + amonth.shift(2)
    # dates = pd.read_csv(crsp_path + 'dates.csv', index_col=0)
    # ret = pd.read_csv(crsp_path + "ret.csv", index_col=0)
    # me = pd.read_csv(crsp_path + 'me.csv', index_col=0)
    # res = runUnivSort(params, ret, indAMO, me, dates)

    print(f"\nCOMPUSTAT derived variables run ended at {datetime.now()}.\n")

    return




import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import statsmodels.api as sm
from AssayingAnomalies.Functions.calcPtfRets import calcPtfRets
from AssayingAnomalies.Functions.estFactorRegs import estFactorRegs
from AssayingAnomalies.Functions.calcTcosts import calcTcosts
from AssayingAnomalies.Functions.prt_sort_results import prt_sort_results

def runUnivSort(params, ret, ind, mcap, dates, **kwargs):

    """
    :param ret:
    :param ind:
    :param mcap:
    :param dates:
    :param kwargs:
        - 'timePeriod' a list or numpy array of dates. can be either start and end date or a vector of dates.
    :return:
    """

    # Convert input DataFrames to Numpy array
    if isinstance(ret, pd.DataFrame):
        ret = ret.to_numpy()

    if isinstance(ind, pd.DataFrame):
        ind = ind.to_numpy()

    if isinstance(mcap, pd.DataFrame):
        mcap = mcap.to_numpy()

    if isinstance(dates, pd.DataFrame):
        dates = dates.values.flatten()

    # Default parameter values
    p = {
        'factorModel': 4,
        'addLongShort': 1,
        'printResults': 1,
        'plotFigure': 1,
        'timePeriod': [dates[0], dates[-1]],
        'holdingPeriod': 1,
        'weighting': 'v',
        'tcosts': -99,
    }

    # Update parameters with user-specified values (if provided).
    for key, value in kwargs.items():
        if key in p:
            p[key] = value

    # Check if weighting is correctly specified
    expected_weighting = {'V', 'v', 'E', 'e'}
    if p['weighting'] not in expected_weighting:
        raise ValueError("Invalid weighting value. Must be one of {'V', 'v', 'E', 'e'}")

    # Update parameters with user-specified values (if provided).
    for key, value in kwargs.items():
        if key in p:
            p[key] = value

    # Check whether the dimensions are correct
    if ret.shape != ind.shape:
        raise ValueError(f"'ret' and 'ind' have different dimensions: {ret.shape} and {ind.shape}.")
    if ret.shape[0] != dates.shape[0]:
        raise ValueError(f"'ret' and 'dates' have different dimensions: {ret.shape} and {dates.shape}.")
    if ret.shape != mcap.shape:
        raise ValueError(f"'ret' and 'mcap' have different dimensions: {ret.shape} and {mcap.shape}.")

    # Check if the factor model code (if entered) is correct
    if isinstance(p['factorModel'], int) and p['factorModel'] not in {1, 3, 4, 5, 6}:
        raise ValueError("Factor model must be 1, 3, 4, 5, 6, or a user-defined matrix.")

    # Check if user entered a subsample
    if p['timePeriod'] != [dates[0], dates[-1]]:
        # Find the start date
        s = next(i for i, date in enumerate(dates) if date >= p['timePeriod'][0])

        # Find the end date, by finding the index 'e' of the last date in the 'dates' array that is <= specified end
        # date in 'p['timePeriod']'. "next()" retrieves first item produced by the generator expression
        if len(p['timePeriod']) == 2:
            # Whatever the user chose
            e = next(i for i, date in reversed(list(enumerate(dates))) if date <= p['timePeriod'][1])
        else:
            # Or the last date in the dates array
            e = len(dates) - 1

        # Subset the input arrays
        ret = ret[s:e + 1, :]
        ind = ind[s:e + 1, :]
        dates = dates[s:e + 1]
        mcap = mcap[s:e + 1, :]

    # Delete all stocks (i.e., columns) that are not held in any portfolio in the full sample
    stock_is_held = np.sum(ind, axis=0) > 0
    ret = ret[:, stock_is_held]
    ind = ind[:, stock_is_held]
    mcap = mcap[:, stock_is_held]

    # Calculate the ptf returns, # stocks, and market caps
    holding_period = p['holdingPeriod']
    weighting = p['weighting'].lower() # returns the user specified weighting as lowercase string
    pret, ptf_num_stocks, ptf_market_cap = calcPtfRets(ret, ind, mcap, holding_period, weighting)

    # Estimate the factor model regressions
    factor_model = p['factorModel']
    add_long_short = p['addLongShort']
    res = estFactorRegs(params=params, pret=pret, dates=dates, factorModel=factor_model, addLongShort=add_long_short)

    # Check if we need to estimate trading costs
    if p['tcosts'] != -99:
        tcosts = p['tcosts']
        if len(tcosts) == 1:
            # If it's just a constant tcost
            tcosts = tcosts * (mcap / mcap)
        else:
            # Remove the stocks that are not held
            tcosts = tcosts[:, stock_is_held]

            # Check if we have to subset
            if 's' in locals():
                tcosts = tcosts[s:e, :]

        # Calculate the actual trading costs
        ptf_costs, ptf_to, dW = calcTcosts(tcosts, ind, mcap, weighting=p['weighting'])

        # Store temporarily the tcosts and net portfolio returns for the
        # long/short portfolio
        tcosts_ts = ptf_costs[:, 0] + ptf_costs[:, -1]
        net_pret = res['pret'][:, -1] - tcosts_ts

        # Regress the net long/short portfolio returns on a constant
        net_res = sm.OLS(100 * net_pret, sm.add_constant(np.ones(net_pret.shape))).fit()

        # Store the tcosts output
        res['tcostsTS'] = tcosts_ts
        res['netpret'] = net_pret
        res['toTS'] = np.column_stack((ptf_to[:, 0], ptf_to[:, -1]))
        res['netxret'] = net_res['beta']
        res['tnetxret'] = net_res['tstat']
        res['turnover'] = np.nanmean(np.nanmean(res['toTS'], axis=0) / 2)
        res['tcosts'] = np.nanmean(res['tcostsTS'])
        res['ptfCosts'] = ptf_costs
        res['ptfTO'] = ptf_to

    # Store a few more variables
    res['w'] = p['weighting']
    res['dates'] = dates
    res['hperiod'] = p['holdingPeriod']
    res['ptfMarketCap'] = ptf_market_cap

    # Add the long-short portfolio time-series # of stocks
    if p['addLongShort'] != 0:
        res['ptfNumStocks'] = np.column_stack((ptf_num_stocks, ptf_num_stocks[:, -1] + ptf_num_stocks[:, 0]))
    else:
        res['ptfNumStocks'] = ptf_num_stocks

    # Print the results
    if p['printResults'] != 0:
        # print(f"p['addLongShort']: {p['addLongShort']}, Type: {type(p['addLongShort'])}")
        prt_sort_results(res, lsprt=p['addLongShort'])

    # Plot the figure
    if p['plotFigure'] != 0 and p['addLongShort'] != 0:
        plt.figure(83)
        plot_strategy_figs(res) # TODO:F Need to make this function

    return res



import pandas as pd
import numpy as np
from scipy.stats import rankdata

def rank_with_nan(data):
    if isinstance(data, pd.DataFrame):
        data = data.to_numpy()

    # Create a copy of the data to avoid modifying the original array
    data_copy = np.array(data, copy=True)

    # Find the NaN elements
    nan_mask = np.isnan(data_copy)

    # Temporarily replace NaNs with a placeholder (max value + 1)
    max_val = np.nanmax(data_copy)
    data_copy[nan_mask] = max_val + 1

    # Rank the data
    ranks = rankdata(data_copy, axis=1)

    # Replace the placeholder ranks with NaN
    ranks[nan_mask] = np.nan

    return ranks


import numpy as np
import pandas as pd
from AssayingAnomalies.Functions.assignToPtf import assignToPtf


def makeUnivSortInd(var, ptfNumThresh, breaksFilterInd=None, portfolioMassInd=None):

    # Validate inputs
    if not (isinstance(var, pd.DataFrame) or isinstance(var, np.ndarray)):
        raise ValueError("Input 'var' must be a DataFrame or a NumPy array.")

    # Convert input DataFrame to Numpy array
    if isinstance(var, pd.DataFrame):
        var = var.values

    if not np.isscalar(ptfNumThresh) and not isinstance(ptfNumThresh, (np.ndarray, list)):
        raise ValueError("Input 'ptfNumThresh' must be a scalar, list, or a NumPy array.")

    if breaksFilterInd is not None:
        if not (isinstance(breaksFilterInd, pd.DataFrame) or isinstance(breaksFilterInd, np.ndarray)):
            raise ValueError("Input 'breaksFilterInd' must be a DataFrame, a NumPy array, or None.")
        if var.shape != breaksFilterInd.shape:
            raise ValueError(f"Input 'breaksFilterInd' has shape {breaksFilterInd.shape} but 'var' has shape {var.shape}.")

    if portfolioMassInd is not None:
        if not (isinstance(portfolioMassInd, pd.DataFrame) or isinstance(portfolioMassInd, np.ndarray)):
            raise ValueError("Input 'portfolioMassInd' must be a DataFrame, a NumPy array, or None.")
        if var.shape != portfolioMassInd.shape:
            raise ValueError("Input 'portfolioMassInd' must have the same shape as 'var'.")

    # If user passed a list instead of numpy array for ptfNumThresh, we will need to change it to numpy array
    if isinstance(ptfNumThresh, list):
        ptfNumThresh = np.array(ptfNumThresh)

    # Determine the breakpoints
    nPtfThresh = len(ptfNumThresh) if isinstance(ptfNumThresh, np.ndarray) else 1

    if nPtfThresh > 1:
        # User entered the breakpoints directly
        bpts = ptfNumThresh
    else:
        # User entered the number of portfolios, so we need to create the breakpoints
        bpts = np.nan * np.empty(ptfNumThresh - 1)
        for i in range(1, ptfNumThresh):
            bpts[i - 1] = i * 100 / ptfNumThresh

    # Check if we are doing cap-weighted breaks
    if portfolioMassInd is None:
        portfolioMassInd = np.ones(var.shape)
    if not np.array_equal(portfolioMassInd, 1):
        # In this case, the user entered a portfolio mass matrix (e.g., market cap)

        portfolioMassInd[np.isnan(var)] = np.nan

        # We'll sort on the variable first & calculate the cumulative market cap
        I = np.argsort(var, axis=1)
        r_ind = np.any(np.isfinite(var), axis=1)
        tempvar = np.empty(var.shape)
        tempvar[:] = np.nan

        for i in np.where(r_ind)[0]:
            temp = portfolioMassInd[i, I[i, :]]
            t_ind = np.isnan(temp)
            temp[t_ind] = 0
            temp = np.cumsum(temp)
            temp = temp / temp[-1]
            temp[t_ind] = np.nan
            tempvar[i, I[i, :]] = temp

        # Assign based on the breakpoints
        bpts = np.concatenate(([0], bpts)) / 100

        ind = np.zeros(var.shape)
        for i in bpts:
            ind = ind + (tempvar > i) # TODO:N Here is causing the RuntimeWarning. Change to np.where(~np.isnan(tempvar), tempvar > i, False)
    else:
        # In this case, we are not doing cap-weighting

        if breaksFilterInd is None:
            breaksFilterInd = np.ones(var.shape)

        # If breaksFilterInd is a matrix, we'll assign nan to observations
        # which we don't want to use for breakpoints (e.g., if breaksFilterInd
        # is an indicator of NYSE stocks, we'll make non-NYSE entries nan)
        if breaksFilterInd is not None:
            breaksFilterInd[breaksFilterInd == 0] = np.nan

        # Calculate the percentiles for var by multiplying by the
        # breaksFilterInd. By default, we'll just multiply by 1. if
        # breaksFilterInd is NYSE, then we only use NYSE stocks for the
        # breakpoints.
        pctileBpointsMat = np.nanpercentile(var * (breaksFilterInd if breaksFilterInd is not None else 1), bpts, axis=1).T

        # Pass the var matrix and the percentile breakpoints matrix to
        # assignToPtf, which returns the indicator matrix for the portfolios
        ind = assignToPtf(var, pctileBpointsMat)

    return ind


# saveFolder = r'/home/jlaws13/PycharmProjects/AssayingAnomalies_root/Data/CRSP/'
# me = pd.read_csv(saveFolder + 'me.csv', index_col=0).astype(float)
# NYSE = pd.read_csv(saveFolder + 'NYSE.csv', index_col=0)
# test_ind = makeUnivSortInd(-me, 10, breaksFilterInd=NYSE)
# test_ind = makeUnivSortInd(-me, [30, 70], breaksFilterInd=NYSE)



import numpy as np
import pandas as pd
import requests
from io import BytesIO
import os
import zipfile


def getFFDailyFactors(params):
    """

    :return:
    """

    # URL to Ken French's daily FF3
    url = 'http://mba.tuck.dartmouth.edu/pages/faculty/ken.french/ftp/F-F_Research_Data_Factors_daily_CSV.zip'

    # Set path to CRSP DailyFolder
    daily_crsp_folder = params.daily_crsp_folder + os.sep

    # Set path to FF Data folder
    FFDataFolder = params.ff_data_folder + os.sep

    # Load dates
    ddates = pd.read_csv(daily_crsp_folder + 'ddates.csv', index_col=0)

    # tricks the host site into thinking the HTTP call is from user instead of machine
    headers = {"User-Agent": "Mozilla/5.0"}
    response = requests.get(url, headers=headers)
    # saving the zip file as an object
    zf = zipfile.ZipFile(BytesIO(response.content))
    # extracting and saving the contents
    zf.extractall(FFDataFolder)
    # loading the csv file that was extracted. the first three rows are useless text so we will skip them
    # :TODO:F Sometimes the file is saved as .csv and other times as .CSV.
    FF3factors = pd.read_csv(FFDataFolder + 'F-F_Research_Data_Factors_daily.CSV', skiprows=3)

    # the final row is also useless so we will remove it
    FF3factors = FF3factors.iloc[:-1, :]

    # Note: The raw data does not have a column for "MKT" factor, instead it is is MKT-RF, but since we didn't add it
    # back in the monthly data I will not add it back now. But the line below that is commented out will correct for
    # this if desired.
    FF3factors.columns = ['dates', 'MKT', 'SMB', 'HML', 'RF']
    # FF3factors['MKT'] = FF3factors['MKT'] + FF3factors['RF']
    # Save the FF daily dates TODO:Q Matlab code saves ffdates as just a copy of the preexisting dates file. But ffdates start at 192607 instead of 192512.
    FF3factors['dates'].to_csv(FFDataFolder + 'ffdates_daily.csv')

    # Intersect our dates with the ones from Ken French Website; ia is index of elements in dates array in common with those
    # in FF.dates and vica versa.
    _, ia, ib = np.intersect1d(np.array(ddates).astype(int), FF3factors['dates'].values.astype(int), return_indices=True)

    #create the variables
    dmkt = np.full(len(ddates), np.nan)
    dsmb = np.full(len(ddates), np.nan)
    dhml = np.full(len(ddates), np.nan)
    drf = np.full(len(ddates), np.nan)

    # first need to change the datatypes from string to float
    FF3factors = FF3factors.astype(float)
    dmkt[ia] = FF3factors.loc[ib, 'MKT'] / 100
    dsmb[ia] = FF3factors.loc[ib, 'SMB'] / 100
    dhml[ia] = FF3factors.loc[ib, 'HML'] / 100
    drf[ia] = FF3factors.loc[ib, 'RF'] / 100

    "Now repeat the steps for UMD factor"
    # URL to Ken French's daily momentum data
    url = 'http://mba.tuck.dartmouth.edu/pages/faculty/ken.french/ftp/F-F_Momentum_Factor_daily_CSV.zip'

    # tricks the host site into thinking the HTTP call is from user instead of machine
    headers = {"User-Agent": "Mozilla/5.0"}
    response = requests.get(url, headers=headers)
    # saving the zip file as an object
    zf = zipfile.ZipFile(BytesIO(response.content))
    # extracting and saving the contents
    zf.extractall(FFDataFolder)
    # loading the csv file that was extracted. the first 12 rows are useless text so we will skip them and the last row
    # is also text so we will remove it
    UMDFactor = pd.read_csv(FFDataFolder + 'F-F_Momentum_Factor_daily.CSV', skiprows=12)
    UMDFactor = UMDFactor.iloc[:-1, :]

    # rename the columns
    UMDFactor.columns = ['dates', 'UMD']

    # Intersect our dates with the ones from Ken French Website; ia is index of elements in dates array in common with those
    # in FF.dates and vica versa.
    _, ia, ib = np.intersect1d(np.array(ddates).astype(int), UMDFactor['dates'].values.astype(int), return_indices=True)

    #create the variables
    dumd = np.full(len(ddates), np.nan)

    # first need to change the datatypes from string to float
    UMDFactor = UMDFactor.astype(float)
    dumd[ia] = UMDFactor.loc[ib, 'UMD'] / 100

    "Now repeat the steps for 5 factors"
    url = 'http://mba.tuck.dartmouth.edu/pages/faculty/ken.french/ftp/F-F_Research_Data_5_Factors_2x3_daily_CSV.zip'

    # tricks the host site into thinking the HTTP call is from user instead of machine
    headers = {"User-Agent": "Mozilla/5.0"}
    response = requests.get(url, headers=headers)
    # saving the zip file as an object
    zf = zipfile.ZipFile(BytesIO(response.content))
    # extracting and saving the contents
    zf.extractall(FFDataFolder)
    # loading the csv file that was extracted. the first 3 rows are useless text so we will skip them
    FF5factors = pd.read_csv(FFDataFolder + 'F-F_Research_Data_5_Factors_2x3_daily.CSV', skiprows=3)
    FF5factors.columns = ['dates', 'MKT', 'SMB', 'HML', 'RMW', 'CMA', 'RF']

    # Intersect our dates with the ones from Ken French Website; ia is index of elements in dates array in common with those
    # in FF.dates and vica versa.
    _, ia, ib = np.intersect1d(np.array(ddates).astype(int), FF5factors['dates'].values.astype(int), return_indices=True)

    #create the variables
    dsmb2 = np.full(len(ddates), np.nan)
    drmw = np.full(len(ddates), np.nan)
    dcma = np.full(len(ddates), np.nan)

    # first need to change the datatypes from string to float
    FF5factors = FF5factors.astype(float)
    dsmb2[ia] = FF5factors.loc[ib, 'SMB'] / 100
    drmw[ia] = FF5factors.loc[ib, 'RMW'] / 100
    dcma[ia] = FF5factors.loc[ib, 'CMA'] / 100

    "Create and save useful dataframes"
    const = np.full(len(drf), 0.01)
    const = pd.Series(const)
    dmkt = pd.Series(dmkt)
    dsmb = pd.Series(dsmb)
    dhml = pd.Series(dhml)
    drf = pd.Series(drf)
    dumd = pd.Series(dumd)
    dsmb2 = pd.Series(dsmb2)
    drmw = pd.Series(drmw)
    dcma = pd.Series(dcma)

    # factors = ['const', 'dmkt', 'dsmb', 'dhml', 'drf', 'dumd', 'dsmb2', 'drmw', 'dcma']

    # :TODO Save ['ffdates', 'const', 'rf', 'mkt', 'smb', 'smb2', 'hml', 'umd', 'rmw', 'cma', 'ff3', 'ff4', 'ff5', 'ff6'] in a giant matrix called ff.csv
    ff = pd.concat([ddates, const, dmkt, dsmb, dhml, drf, dumd, dsmb2, drmw, dcma], axis=1)
    ff.columns = ['dates', 'const', 'mkt', 'smb', 'hml', 'rf', 'umd', 'smb2', 'rmw', 'cma']
    ff.to_csv(FFDataFolder + 'dff.csv')

    ff3 = pd.concat([ddates, const, dmkt, dsmb, dhml], axis=1)
    ff3.columns = ['dates', 'const', 'mkt', 'smb', 'hml']
    ff3.to_csv(FFDataFolder + 'dff3.csv')

    ff4 = pd.concat([ff3, dumd], axis=1)
    ff4.columns = ['ddates', 'const', 'dmkt', 'dsmb', 'dhml', 'dumd']
    ff4.to_csv(FFDataFolder + 'dff4.csv')

    ff5 = pd.concat([ddates, const, dmkt, dsmb2, dhml, drmw, dcma], axis=1)
    ff5.columns = ['dates', 'const', 'mkt', 'smb2', 'hml', 'rmw', 'cma']
    ff5.to_csv(FFDataFolder + 'dff5.csv')

    ff6 = pd.concat([ff5, dumd], axis=1)
    ff6.columns = ['dates', 'const', 'mkt', 'smb2', 'hml', 'rmw', 'cma', 'umd']
    ff6.to_csv(FFDataFolder + 'dff6.csv')

    return

# getFFDailyFactors(params)


import numpy as np
import pandas as pd
from datetime import datetime
import os
import statsmodels.api as sm
from multiprocessing import Pool
import dask.dataframe as dd
from .getFFDailyFactors import getFFDailyFactors


# -Deprecated
# def parallel_compute_ivol(i, j, col_index, ind1, ind3, dret, dmkt, dff3):
#     # Regression for 1-month residual
#     X = sm.add_constant(dmkt[ind1])  # Add a constant term
#     y = dret.loc[ind1, j]
#     model = sm.OLS(y.values, X.values, missing='drop')
#     results = model.fit()
#     ivol = np.sqrt(np.mean(results.resid ** 2))
#
#     # Regression for FF3 1-month residual
#     X_ff3 = dff3.loc[ind1]
#     y = dret.loc[ind1, j]
#     model_ff3 = sm.OLS(y.values, X_ff3.values, missing='drop')
#     results_ff3 = model_ff3.fit()
#     iffvol = np.sqrt(np.mean(results_ff3.resid ** 2))
#
#     # Regression for 3-month residual
#     X = sm.add_constant(dmkt[ind3])  # Add a constant term
#     y = dret.loc[ind3, j]
#     model = sm.OLS(y.values, X.values, missing='drop')
#     results = model.fit()
#     ivol3 = np.sqrt(np.mean(results.resid ** 2))
#
#     # Regression for FF3 3-month residual
#     X_ff3 = dff3.loc[ind3]
#     y = dret.loc[ind3, j]
#     model_ff3 = sm.OLS(y.values, X_ff3.values, missing='drop')
#     results_ff3 = model_ff3.fit()
#     iffvol3 = np.sqrt(np.mean(results_ff3.resid ** 2))
#
#     return i, col_index, ivol, iffvol, ivol3, iffvol3


def parallel_compute_ivol(i, col_index, y1, X1, y3, X3, Xf1, Xf3):
    # Regression for 1-month residual
    model = sm.OLS(y1, X1, missing='drop')
    results = model.fit()
    ivol = np.sqrt(np.mean(results.resid ** 2))

    # Regression for FF3 1-month residual
    model_ff3 = sm.OLS(y1, Xf1, missing='drop')
    results_ff3 = model_ff3.fit()
    iffvol = np.sqrt(np.mean(results_ff3.resid ** 2))

    # Regression for 3-month residual
    model = sm.OLS(y3, X3, missing='drop')
    results = model.fit()
    ivol3 = np.sqrt(np.mean(results.resid ** 2))

    # Regression for FF3 3-month residual
    model_ff3 = sm.OLS(y3, Xf3, missing='drop')
    results_ff3 = model_ff3.fit()
    iffvol3 = np.sqrt(np.mean(results_ff3.resid ** 2))

    return i, col_index, ivol, iffvol, ivol3, iffvol3


def makeCRSPDailyDerivedVariables(params):
    # Timekeeping
    print(f"\nNow working on making CRSP daily derived variables. Run started at {datetime.now()}.\n")

    # set the path to crsp daily folder
    crsp_folder = params.crspFolder + os.sep

    # set the path to crsp daily folder
    daily_crsp_folder = params.daily_crsp_folder + os.sep

    # set the path to ff data folder
    ff_data_folder = params.ff_data_folder + os.sep

    "Make market capitalization"
    dme = np.abs(dd.read_parquet(daily_crsp_folder + 'dprc.parquet')) \
          * dd.read_parquet(daily_crsp_folder + 'dshrout.parquet') / 1000
    dme = dme.where(dme != 0, np.nan)
    dme = dme.compute()
    dme.to_parquet(daily_crsp_folder + 'dme.parquet')

    "Load the necessary data from storage"
    dret_x_dl = dd.read_parquet(daily_crsp_folder + 'dret_x_adj.parquet').compute()
    permno = pd.read_csv(crsp_folder + 'permno.csv', index_col=0)
    crsp_dsedelist = pd.read_csv(daily_crsp_folder + 'crsp_dsedelist.csv', index_col=0)

    "get permnos to drop"
    #get the indices of the permnos in crsp_msedelist.permno  that are NOT in permno
    mask = np.isin(crsp_dsedelist.permno, permno, invert=True)
    # get the permnos to remove
    toDrop = crsp_dsedelist.permno[mask]

    "drop these permnos from crsp_dsedelist dataframe"
    crsp_dsedelist = crsp_dsedelist.set_index('permno')
    crsp_dsedelist = crsp_dsedelist.drop(toDrop)
    #changing permnos back to a column instead of index
    crsp_dsedelist.reset_index(inplace=True)

    # The index of dret_x_dl is 'YYYY-MM-DD' but ddates is saved as 'YYYYMMDD', therefor I recreate the series to match.
    ddates = dd.read_parquet(daily_crsp_folder + 'crsp_dsf*.parquet', columns='date').unique().astype(str)
    ddates = ddates.compute()

    "fill in the delisting returns"
    dret = dret_x_dl.copy()
    for i in range(len(crsp_dsedelist)):
        # for i in range(500):
        "find the index in permno of ith permno in the delist dataframe"
        c = np.where(permno == crsp_dsedelist['permno'][i])[0][0]
        # print(c)
        "find the index in dates that corresponds to the delisting date of the ith permno"
        r_dt_ar = np.where(ddates == crsp_dsedelist['dlstdt'][i])[0]
        if r_dt_ar.size == 0:
            continue
        else:
            r_dt = r_dt_ar[0]
        # print(r_dt)
        "returns a series of row numbers in ret dataframe where the delisted permno has finite returns"
        r_last_array = np.where(np.isfinite(dret.iloc[:, c]) == True)[0]
        "check to see of the array is empty, and if it is, skip to the next iteration of the loop"
        if r_last_array.size == 0:
            continue
        else:
            "selects the last index, i.e., the row number of the final finite return for delisted permno"
            r_last = r_last_array[-1] + 1
            # print(r_last)
        if np.isfinite(dret.iloc[r_dt, c]):
            if r_last < len(ddates):
                "assigns the delisting return to the position following the last finite return"
                dret.iloc[(r_last), c] = crsp_dsedelist.dlret[i]
        else:
            dret.iloc[r_dt, c] = crsp_dsedelist.dlret[i]

    "Save the returns dataframe"
    dret.columns = dret.columns.astype(str)
    dret.to_parquet(daily_crsp_folder + 'dret.parquet')

    "Adjust NASDAQ volume following Gao and Ritter (2010). See their Appendix B for more details"
    "To adjust the volume, we first need to create the daily exchange code matrix"
    exchcd = pd.read_csv(crsp_folder + 'exchcd.csv', index_col=0)
    exchcd.index = pd.to_datetime(exchcd.index.astype(int), format='%Y%m')
    ddates = pd.read_csv(daily_crsp_folder + 'ddates.csv', index_col=0).values.flatten().astype(int)
    ddates = pd.to_datetime(ddates, format='%Y%m%d')
    dexchcd = exchcd.reindex(ddates, method='ffill')
    dexchcd.to_parquet(daily_crsp_folder + 'dexchcd.parquet')

    # Slice dexchcd to match the size of dvol
    dvol = dd.read_parquet(daily_crsp_folder + 'dvol_x_adj.parquet').compute()
    dvol_columns = dd.read_parquet(daily_crsp_folder + 'dvol_x_adj.parquet').columns

    # Note: The adjusted_dvol computation assumes dvol data aligns exactly with dexchcd in structure and partitioning
    "Requires too much memory to do entire dataframe at once. Will try and go column by column."
    # Divide by 2 prior to Feb 2001
    # Divide by 1.8 for most of 2001
    # Divide by 1.6 for 2002 and 2003
    columns_to_process = dexchcd.columns[(dexchcd == 3).any()].to_list()
    for column in columns_to_process:
        print(column)
        col_index = dvol_columns.get_loc(column)
        # Since iterating over columns, I only need the row numbers where both conditions are satisfied for a particular column
        rows_first = np.where((dexchcd[column] == 3) & (dexchcd.index < pd.Timestamp('2001-02-01')))[0]
        rows_mid = np.where((dexchcd[column] == 3) & (dexchcd.index >= pd.Timestamp('20010201')) & (dexchcd.index < pd.Timestamp('20020101')))[0]
        rows_last = np.where((dexchcd[column] == 3) & (dexchcd.index >= pd.Timestamp('20020101')) & (dexchcd.index < pd.Timestamp('20040101')))[0]

        dvol.iloc[rows_first, col_index] = dvol.iloc[rows_first, col_index] / 2
        dvol.iloc[rows_mid, col_index] = dvol.iloc[rows_mid, col_index] / 1.8
        dvol.iloc[rows_last, col_index] = dvol.iloc[rows_last, col_index] / 1.6

    dvol.columns = dvol.columns.astype(str)
    dvol.to_parquet(daily_crsp_folder + 'dvol.parquet')

    "Download, clean up, and save the Fama-French factors from Ken French's website"
    getFFDailyFactors(params=params)

    "Make Amihud values"
    # first 12 month rolling amihud values

    # load the monthly returns, monthly price, and monthly dates, matrices
    ret = pd.read_csv(crsp_folder + 'ret.csv', index_col=0).astype(float)
    prc = pd.read_csv(crsp_folder + 'prc.csv', index_col=0).astype(float)
    dates = pd.read_csv(crsp_folder + 'dates.csv', index_col=0).astype(int)

    # replace zero-volume observations with nan and calculate dollar volume
    # dprc = pd.read_parquet(daily_crsp_folder + 'dprc.parquet')
    dprc = dd.read_parquet(daily_crsp_folder + 'dprc.parquet').compute()
    dollar_volume = np.abs(dprc) * np.abs(dvol)
    zeroVolInd = (dollar_volume == 0)
    dollar_volume[zeroVolInd] = np.nan

    # calculate the daily price impact
    dret = dd.read_parquet(daily_crsp_folder + 'dret.parquet').compute().astype(float)
    price_impact = np.divide(np.abs(dret), dollar_volume)

    # Take the absolute value of the monthly price and store the number of months
    prc = np.abs(prc)
    nmonths = len(ret.index)

    # Reload ddates in YYYYMMDD format
    ddates = pd.read_csv(daily_crsp_folder + 'ddates.csv', index_col=0).astype(int)

    # initialize amihud matrix
    amihud = np.full_like(ret, np.nan)

    # Calculate amihud
    for i in range(11, nmonths):
        print(f"Month {i}/{nmonths}")
        # Find the days in the last year
        start_date = dates.iloc[i - 11]
        end_date = dates.iloc[i]
        index_year = ((ddates // 100 >= start_date) & (ddates // 100 <= end_date))

        # Ensure index_year is a 1D boolean array
        index_year = index_year.values.flatten()

        # Store monthly price impact for last year
        last_yr_pi = price_impact[index_year].copy()

        # Apply the filters (200 obs & price > $5)
        idx_to_drop = np.isfinite(last_yr_pi).sum(axis=0) < 200
        idx_to_drop |= prc.iloc[i, :].values.flatten() <= 5
        last_yr_pi.iloc[:, idx_to_drop] = np.nan

        # Calculate the mean price impact for the last year, ignoring NaN values
        amihud[i, :] = np.nanmean(last_yr_pi, axis=0)

    amihud = pd.DataFrame(amihud)
    amihud.index = dates.values.flatten()
    amihud.columns = permno.values.flatten().astype(str)
    amihud.to_parquet(crsp_folder + 'amihud.parquet')

    "Make other measures from the daily data"
    # Realized volatilities
    # Initialize the matrices
    n_months, n_stocks = ret.shape
    RVOL_matrices = {period: np.full((n_months, n_stocks), np.nan) for period in [1, 3, 6, 12, 36, 60]}
    dshvol = np.full_like(ret, np.nan)
    dshvolM = np.full_like(ret, np.nan)
    dretmax = np.full_like(ret, np.nan)
    dretmin = np.full_like(ret, np.nan)

    # First turn the dates and ddates dataframes into arrays
    ddates_array = ddates.values.flatten()
    dates_array = dates.values.flatten()

    # Find the first month index
    first_month_date = ddates_array[0] // 100
    if first_month_date in dates_array:
        first_month = np.where(dates_array == first_month_date)[0][0]
    else:
        raise ValueError("First month date not found in dates array")

    # Monthly loop range
    last_month = len(dates)

    # Calculate measures
    for i in range(first_month, last_month):
        for period, RVOL in RVOL_matrices.items():
            start_index = max(i - period + 1, 0)
            ind = (ddates_array // 100 >= dates_array[start_index]) & (ddates_array // 100 <= dates_array[i])

            # Ensure ind is a 1D boolean array and matches the number of rows in dret
            ind = ind.flatten()

            # Apply the indexing to the DataFrame
            selected_dret = dret.loc[ind, :]

            RVOL[i, :] = np.nanstd(selected_dret, axis=0)

        ind1 = (ddates_array // 100 == dates_array[i])

        # Ensure ind1 is a 1D boolean array and matches the number of rows in dvol
        ind1 = ind1.flatten()

        # Apply the indexing to the DataFrame
        selected_dvol = dvol.loc[ind1, :]

        dshvol[i, :] = np.nansum(selected_dvol, axis=0)
        dshvolM[i, :] = np.nanmax(selected_dvol, axis=0)
        dretmax[i, :] = np.nanmax(selected_dvol, axis=0)
        dretmin[i, :] = np.nanmin(selected_dvol, axis=0)

    # Save the matrices
    for period, RVOL_matrix in RVOL_matrices.items():
        filename = f'RVOL{period}.csv'
        RVOL_matrix = pd.DataFrame(RVOL_matrix)
        RVOL_matrix.to_csv(crsp_folder + filename)

    dshvol = pd.DataFrame(dshvol)
    dshvol.columns = dshvol.columns.astype(str)
    dshvol.to_parquet(crsp_folder + 'dshvol.parquet')

    dshvolM = pd.DataFrame(dshvolM)
    dshvolM.columns = dshvolM.columns.astype(str)
    dshvolM.to_parquet(crsp_folder + 'dshvolM.parquet')

    dretmax = pd.DataFrame(dretmax)
    dretmax.columns = dretmax.columns.astype(str)
    dretmax.to_parquet(crsp_folder + 'dretmax.parquet')

    dretmin = pd.DataFrame(dretmin)
    dretmin.columns = dretmin.columns.astype(str)
    dretmin.to_parquet(crsp_folder + 'dretmin.parquet')

    "Make IVOLs"
    # Load the FF factors
    dff = pd.read_csv(ff_data_folder + 'dff.csv', index_col=0).astype(float)
    dff3 = pd.read_csv(ff_data_folder + 'dff3.csv', index_col=0).astype(float)
    dff3.index = dff3['dates']
    dff3.drop(columns=['dates'], inplace=True)
    dmkt = dff['mkt'].copy()

    # Create the daily excess returns above Rf
    drf = dff['rf'].to_numpy()
    rf_mat = np.tile(drf, (len(ret.columns), 1)).T
    dxret = dret - rf_mat

    # Initialize the IVOL variables
    ivol = np.full_like(ret, np.nan)
    ivol3 = np.full_like(ret, np.nan)
    iffvol = np.full_like(ret, np.nan)
    iffvol3 = np.full_like(ret, np.nan)

    if params.remote_or_not:
        total_months = last_month - first_month
        batches = 10
        batch_size = max(total_months // batches, 1)
        for batch in range(batches):
            start_i = first_month + batch*batch_size
            end_i = min(start_i + batch_size, last_month)
            # Prepare a list of arguments for parallel processing
            arguments = []
            # for i in range(first_month, last_month):
            for i in range(start_i, end_i):
                print(i)
                ind1 = ((ddates.values.flatten() // 100) == dates.iloc[i, 0])
                ind3 = ((ddates.values.flatten() // 100) >= dates.iloc[max(i - 2, 0), 0]) & (
                        (ddates.values.flatten() // 100) <= dates.iloc[i, 0])
                hor_ind = np.isfinite(dxret.loc[ind1]).sum(axis=0) > 0
                X1 = sm.add_constant(dmkt[ind1].values)
                X1f = sm.add_constant(dff3.loc[ind1].values)
                X3 = sm.add_constant(dmkt[ind3].values)
                X3f = sm.add_constant(dff3.loc[ind3].values)

                for j in hor_ind[hor_ind].index:
                # for j in hor_ind[hor_ind].index[:5]:
                    col_index = dret.columns.get_loc(j)
                    y1 = dret.loc[ind1, j].values
                    y3 = dret.loc[ind3, j].values

                    arguments.append((i, col_index, y1, X1, y3, X3, X1f, X3f))

            # Run the computation in parallel
            with Pool(processes=params.num_cpus) as pool:
                results = pool.starmap(parallel_compute_ivol, arguments)

            # Unpack results and populate the matrices
            for res in results:
                i, col_index, ivol_val, iffvol_val, ivol3_val, iffvol3_val = res
                ivol[i, col_index] = ivol_val
                iffvol[i, col_index] = iffvol_val
                ivol3[i, col_index] = ivol3_val
                iffvol3[i, col_index] = iffvol3_val

    else:
        # Monthly loop range
        first_month_date = ddates.iloc[0, 0] // 100
        first_month = np.where(dates.values.flatten() == first_month_date)[0][0]
        last_month = len(dates)

        # Calculate IVOLs
        for i in range(first_month, last_month):
            # Find the 1- and 3-month daily indices
            ind1 = ((ddates.values.flatten() // 100) == dates.iloc[i, 0])
            ind3 = ((ddates.values.flatten() // 100) >= dates.iloc[max(i - 2, 0), 0]) & (
                    (ddates.values.flatten() // 100) <= dates.iloc[i, 0])

            # Stocks which we need to loop through
            hor_ind = dxret.iloc[ind1, :].apply(lambda x: np.isfinite(x).sum(), axis=0) > 0

            for j in hor_ind[hor_ind].index:
                # Get the integer index for the column
                col_index = dret.columns.get_loc(j)

                # Regression for 1-month residual
                X = sm.add_constant(dmkt[ind1])  # Add a constant term
                y = dret.loc[ind1, j]
                model = sm.OLS(y.values, X.values, missing='drop')
                results = model.fit()
                ivol[i, col_index] = np.sqrt(np.mean(results.resid ** 2))

                # Regression for FF3 1-month residual
                X_ff3 = dff3.loc[ind1]
                y = dret.loc[ind1, j]
                model_ff3 = sm.OLS(y.values, X_ff3.values, missing='drop')
                results_ff3 = model_ff3.fit()
                iffvol[i, col_index] = np.sqrt(np.mean(results_ff3.resid ** 2))

                # Regression for 3-month residual
                X = sm.add_constant(dmkt[ind3])  # Add a constant term
                y = dret.loc[ind3, j]
                model = sm.OLS(y.values, X.values, missing='drop')
                results = model.fit()
                ivol3[i, col_index] = np.sqrt(np.mean(results.resid ** 2))

                # Regression for FF3 3-month residual
                X_ff3 = dff3.loc[ind3]
                y = dret.loc[ind3, j]
                model_ff3 = sm.OLS(y.values, X_ff3.values, missing='drop')
                results_ff3 = model_ff3.fit()
                iffvol3[i, col_index] = np.sqrt(np.mean(results_ff3.resid ** 2))

    # Save the IVOLs as parquet files
    ivol = pd.DataFrame(ivol)
    ivol.columns = ivol.columns.astype(str)
    ivol.to_parquet(crsp_folder + 'IVOL.parquet')

    iffvol = pd.DataFrame(iffvol)
    iffvol.columns = iffvol.columns.astype(str)
    iffvol.to_parquet(crsp_folder + 'IffVOL.parquet')

    ivol3 = pd.DataFrame(ivol3)
    ivol3.columns = ivol3.columns.astype(str)
    ivol3.to_parquet(crsp_folder + 'IVOL3.parquet')

    iffvol3 = pd.DataFrame(iffvol3)
    iffvol3.columns = iffvol3.columns.astype(str)
    iffvol3.to_parquet(crsp_folder + 'IffVOL3.parquet')

    "Make CAR3s: Inilialize the CAR3 matrix, calculate abnormal returns, iterate through each month to compute the" \
    "CAR3 value for each stock at each announcement date"
    # load additional variables
    rdq = pd.read_csv(params.compFolder + os.sep + 'RDQ.csv', index_col=0)
    fqtr = pd.read_csv(params.compFolder + os.sep + 'FQTR.csv', index_col=0)
    # nyse = pd.read_csv(params.crspFolder + os.sep + 'NYSE.csv', index_col=0)
    me = pd.read_csv(params.crspFolder + os.sep + 'me.csv', index_col=0)

    # For some reason RDQ and FQTR has an additional row at the beginning of the matrix. It doesn't correspond to any
    # date so I will check if this is the case and remove it.
    if pd.isna(rdq.index[0]):
        rdq = rdq.iloc[1:, :]

    if pd.isna(fqtr.index[0]):
        fqtr = fqtr.iloc[1:, :]

    # Only leave the announcement dates
    rdq = rdq.where(~fqtr.isna(), np.nan)

    # Initialize the CAR3 matrix
    CAR3 = np.full_like(rdq, np.nan)

    # Calculate the abnormal return (in excess of the market); (ret_{i,t} - ret_{mkt,t})
    rptdMkt_values = (1 + dmkt + drf).values
    rptdMkt = np.tile(rptdMkt_values, (len(me.columns), 1)).T
    dxret = 1 + dret - rptdMkt

    # Monthly loop range
    first_month = np.where(rdq.notna().sum(axis=1) > 0)[0][0]
    last_month = len(dates)

    # Loop over the months
    for i in range(first_month, last_month):
        # Find all the announcements in the current month
        idx_announcements = np.where(rdq.iloc[i, :].notna())[0]

        # Loop over the announcements
        for c in idx_announcements:
            permno = rdq.columns[c]  # Get the permno (column label) from rdq

            # Check if the permno is in the columns of dxret
            if permno in dxret.columns:
                dxret_col_index = dxret.columns.get_loc(permno)  # Find the corresponding column in dxret

                # Get the announcement date and find the corresponding row in ddates
                announcement_date = rdq.iloc[i, c]
                r = np.where(ddates == announcement_date)[0]

                # Calculate the CAR3 if we found a match
                if r.size > 0 and r[0] < len(ddates) - 1:
                    CAR3[i, c] = np.prod(dxret.iloc[r[0] - 1:r[0] + 2, dxret_col_index]) - 1

    # Fill the observations between quarters; Matlab uses a helper function, but this isn't necessary in Python
    CAR3 = pd.DataFrame(CAR3).ffill(axis=0)

    # Change the columns names to be integers
    ret.columns = [int(float(col)) for col in ret.columns]
    CAR3.columns = [int(float(col)) for col in rdq.columns]

    # Identify common columns between 'ret' and 'CAR3'
    common_columns = ret.columns.intersection(CAR3.columns)

    # Initialize a mask with False for all entries in CAR3
    idx_to_drop = pd.DataFrame(False, index=ret.index, columns=CAR3.columns)

    # Update the mask for common columns based on NaN values in 'ret'
    idx_to_drop.loc[:, common_columns] = ret[common_columns].isna()

    # Remove all observations where ret is nan
    CAR3[idx_to_drop] = np.nan

    # Save
    # CAR3.to_csv(crsp_folder + 'CAR3.csv')
    CAR3.columns = CAR3.columns.astype(str)
    CAR3.to_parquet(crsp_folder + 'CAR3.parquet')

    "Timekeeping"
    print(f"\nCRSP daily derived variables run ended at {datetime.now()}.\n")

    return


# makeCRSPDailyDerivedVariables(params=params)





import pandas as pd
import numpy as np
import os


def makeGibbs(params, file_path):
    """
    This function processes and manipulates financial data to create a 'Gibbs' matrix. It reads in stored estimates from
     a given file path, cleans the data, and computes yearly averages for each 'permno' and 'year' combination. The
     function also merges this data with a CRSP link table, reshapes it, and then performs a final transformation by
     multiplying the values by 2.

    Parameters:
    ----------
    params : object
        An object containing attributes related to data paths:
        - crspFolder: str
            The file path to the CRSP data folder.
    file_path : str
        The file path to the CSV file containing the estimates to be processed.

    Process:
    --------
    1. Read and clean the input data from 'file_path'.
    2. Calculate yearly averages for each 'permno' and 'year'.
    3. Merge with the CRSP link table.
    4. Reshape the merged data into a Gibbs matrix.
    5. Multiply the matrix by 2.

    Returns:
    -------
    gibbs : numpy.ndarray
        A numpy array representing the Gibbs matrix, which contains the processed and transformed data.

    Notes:
    ------
    - The input file should contain columns 'permno', 'year', and 'c'.
    - The Gibbs matrix is derived by merging CRSP link data and averaging over years, then reshaping and scaling.
    """

    # set path to crsp folder
    crsp_path = params.crspFolder + os.sep

    # read in the stored estimates
    # data = pd.read_csv(crsp_path + filename, index_col=0)
    # data = data[['permno', 'year', 'c']]
    data = pd.read_csv(file_path, usecols=['permno', 'year', 'c'])

    # Clean them up:
    data = data.dropna(subset=['c'])

    # The Gibbs file is monthly so in some years there is more than one estimate. We will take the yearly average.
    # First group by 'permno' and 'year' and if there are multiple entries then take the average of them. The
    # 'reset_index()' call is used to convert the indices created by 'groupby() back into columns, which makes the
    # DataFrame look more like a regular table.'
    data = data.groupby(['permno', 'year']).mean().reset_index()
    data[['permno', 'year']] = data[['permno', 'year']].astype(int)

    # Load the CRSP link table
    crsp_link = pd.read_csv(crsp_path + 'crsp_link.csv', index_col=0)

    # Create a year variable
    crsp_link['date'] = crsp_link['date'].str.replace('-', '').astype(int)
    crsp_link['year'] = crsp_link['date'].astype(int) // 10000

    # Merge the data
    # Ensure the key columns ('permno', 'year' in this case) are of the same data type in both DataFrames
    mergedData = pd.merge(crsp_link, data, how='left', on=['permno', 'year'])

    # Reshape the data using unstack
    # This step might need adjustment based on the structure of your 'mergedData' and what you want to achieve with 'unstack'
    gibbs = mergedData.pivot(index='date', columns='permno', values='c')

    # Convert the DataFrame to a numpy array and transpose it if necessary
    gibbs = gibbs.to_numpy()

    # Multiply by 2
    gibbs *= 2

    return pd.DataFrame(gibbs)



import numpy as np
import pandas as pd

def makeFF10Indus(SIC):
    """
    PURPOSE: This function creates a Fama and French 10-industry classification index and a table with the industry
             names.
    :param SIC: Dataframe containing industry classifications, e.g. siccd
    :return:
            - FF10: dataframe with the industries indicators
            - FF10Names: table with the industry names
    """
    # 1 NoDur Consumer NonDurables -- Food, Tobacco, Textiles, Apparel, Leather, Toys
    FF10 = ((SIC >= 100) & (SIC <= 999)).astype(int)
    FF10 += ((SIC >= 2000) & (SIC <= 2399)).astype(int)
    FF10 += ((SIC >= 2700) & (SIC <= 2749)).astype(int)
    FF10 += ((SIC >= 2770) & (SIC <= 2799)).astype(int)
    FF10 += ((SIC >= 3100) & (SIC <= 3199)).astype(int)
    FF10 += ((SIC >= 3940) & (SIC <= 3989)).astype(int)

    # 2 Durbl Consumer Durables -- Cars, TV's, Furniture, Household Appliances
    FF10 += 2 * (((SIC >= 2500) & (SIC <= 2519)).astype(int))
    FF10 += 2 * (((SIC >= 2590) & (SIC <= 2599)).astype(int))
    FF10 += 2 * (((SIC >= 3630) & (SIC <= 3659)).astype(int))
    FF10 += 2 * (((SIC >= 3710) & (SIC <= 3711)).astype(int))
    FF10 += 2 * (((SIC == 3714)).astype(int))
    FF10 += 2 * (((SIC == 3716)).astype(int))
    FF10 += 2 * (((SIC >= 3750) & (SIC <= 3751)).astype(int))
    FF10 += 2 * (((SIC == 3792)).astype(int))
    FF10 += 2 * (((SIC >= 3900) & (SIC <= 3939)).astype(int))
    FF10 += 2 * (((SIC >= 3990) & (SIC <= 3999)).astype(int))

    # 3 Manuf Manufacturing -- Machinery, Trucks, Planes, Chemicals, Off Furn, Paper, Com Printing
    FF10 += 3 * (((SIC >= 2520) & (SIC <= 2589)).astype(int))
    FF10 += 3 * (((SIC >= 2600) & (SIC <= 2699)).astype(int))
    FF10 += 3 * (((SIC >= 2750) & (SIC <= 2769)).astype(int))
    FF10 += 3 * (((SIC >= 2800) & (SIC <= 2829)).astype(int))
    FF10 += 3 * (((SIC >= 3000) & (SIC <= 3099)).astype(int))
    FF10 += 3 * (((SIC >= 3200) & (SIC <= 3569)).astype(int))
    FF10 += 3 * (((SIC >= 3580) & (SIC <= 3621)).astype(int))
    FF10 += 3 * (((SIC >= 3623) & (SIC <= 3629)).astype(int))
    FF10 += 3 * (((SIC >= 3700) & (SIC <= 3709)).astype(int))
    FF10 += 3 * (((SIC >= 3712) & (SIC <= 3713)).astype(int))
    FF10 += 3 * (((SIC == 3715)).astype(int))
    FF10 += 3 * (((SIC >= 3717) & (SIC <= 3749)).astype(int))
    FF10 += 3 * (((SIC >= 3752) & (SIC <= 3791)).astype(int))
    FF10 += 3 * (((SIC >= 3793) & (SIC <= 3799)).astype(int))
    FF10 += 3 * (((SIC >= 3860) & (SIC <= 3899)).astype(int))

    # 4 Enrgy Oil, Gas, and Coal Extraction and Products
    FF10 += 4 * (((SIC >= 1200) & (SIC <= 1399)).astype(int))
    FF10 += 4 * (((SIC >= 2900) & (SIC <= 2999)).astype(int))

    #  5 HiTec  Business Equipment -- Computers, Software, and Electronic Equipment
    FF10 += 5*((SIC >=3570) & (SIC <= 3579))
    FF10 += 5*((SIC >=3622) & (SIC <= 3622))  # Industrial controls
    FF10 += 5*((SIC >=3660) & (SIC <= 3692))
    FF10 += 5*((SIC >=3694) & (SIC <= 3699))
    FF10 += 5*((SIC >=3810) & (SIC <= 3839))
    FF10 += 5*((SIC >=7370) & (SIC <= 7372))  # Services - computer programming and data processing
    FF10 += 5*((SIC >=7373) & (SIC <= 7373))  # Computer integrated systems design
    FF10 += 5*((SIC >=7374) & (SIC <= 7374))  # Services - computer processing, data prep
    FF10 += 5*((SIC >=7375) & (SIC <= 7375))  # Services - information retrieval services
    FF10 += 5*((SIC >=7376) & (SIC <= 7376))  # Services - computer facilities management service
    FF10 += 5*((SIC >=7377) & (SIC <= 7377))  # Services - computer rental and leasing
    FF10 += 5*((SIC >=7378) & (SIC <= 7378))  # Services - computer maintanence and repair
    FF10 += 5*((SIC >=7379) & (SIC <= 7379))  # Services - computer related services
    FF10 += 5*((SIC >=7391) & (SIC <= 7391))  # Services - R&D labs
    FF10 += 5*((SIC >=8730) & (SIC <= 8734))  # Services - research, development, testing labs


    # 6 Telcm Telephone and Television Transmission
    FF10 += 6*((SIC >=4800) & (SIC <= 4899))

    # 7 Shops Wholesale, Retail, and Some Services (Laundries, Repair Shops)
    FF10 += 7*((SIC >=5000) & (SIC <= 5999))
    FF10 += 7*((SIC >=7200) & (SIC <= 7299))
    FF10 += 7*((SIC >=7600) & (SIC <= 7699))

    # 8 Hlth Healthcare, Medical Equipment, and Drugs
    FF10 += 8*((SIC >=2830) & (SIC <= 2839))
    FF10 += 8*((SIC >=3693) & (SIC <= 3693))
    FF10 += 8*((SIC >=3840) & (SIC <= 3859))
    FF10 += 8*((SIC >=8000) & (SIC <= 8099))

    # 9 Utils Utilities
    FF10 += 9*((SIC >=4900) & (SIC <= 4949))

    # 10 Other Other -- Mines, Constr, BldMt, Trans, Hotels, Bus Serv, Entertainment, Finance
    FF10[(SIC > 0) & (FF10 == 0)] = 10

    # Store the FF10 industries and their names
    FF10Names = pd.DataFrame({
        'number': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        'shortName': ['NoDur', 'Durbl', 'Manuf', 'Enrgy', 'HiTec', 'Telcm', 'Shops', 'Hlth', 'Utils', 'Other'],
        'longName': ['Consumer NonDurables', 'Consumer Durables', 'Manufacturing',
                     'Oil, Gas, and Coal Extraction and Products', 'Business Equipment',
                     'Telephone and Television Transmission', 'Wholesale, Retail, and Some Services',
                     'Healthcare, Medical Equipment, and Drugs', 'Utilities', 'Other']
    })
    FF10Names = FF10Names.set_index('number')

    return FF10, FF10Names

# test_indus, test_names = makeFF10Indus(SIC)






import numpy as np

# :TODO:F the spacing between column labels is off.
def prt_sort_results(res, lsprt=1):
    # print(f"lsprt: {lsprt}, Type: {type(lsprt)}")
    """
    Utility function to optionally print results by runUnivSort().

    Parameters:
    ----------
    res : dict
        A dictionary containing the results returned by runUnivSort().

    lsprt : int, optional
        A flag to indicate whether to print a long/short portfolio or not.
    """

    if res['w'].lower() == 'e':
        www = 'Equally-weighted '
    elif res['w'].lower() == 'v':
        www = 'Value-weighted'

    if isinstance(res['factorModel'], int):
        mdl = f"Fama and French {res['nFactors']}-factor model"
        if res['factorModel'] == 1:
            mdl = 'CAPM'
    else:
        mdl = f"User-defined {res['nFactors']}-factor model"

    heads = [label for label in res['factorLabels']]
    X = np.column_stack((res['xret'], res['alpha'], res['factorLoadings']))
    T = np.column_stack((res['txret'], res['talpha'], res['tfactors']))

    fmtb = "{:7.3f}"
    fmt = "[{:5.2f}]"
    S1 = " "  # one spaces
    S2 = "  "  # two spaces
    max_col_widths = [len(fmtb.format(val)) + 2 * len(S2) for val in X[0, :]]
    B = []
    for j in range(X.shape[0]):
        A = []
        A_t = []
        for i in range(X.shape[1]):
            coeff_space = " " * (max_col_widths[i] - len(fmtb.format(X[j, i])) - 2)
            tstat_space = " " * (max_col_widths[i] - len(fmt.format(T[j, i])) - 2)
            A.append(f"{S2}{fmtb.format(X[j, i])}{coeff_space}")
            A_t.append(f"{S2}{fmt.format(T[j, i])}{tstat_space}")

        B.append(S1.join(A))
        B.append(S1.join(A_t))

    Q = []
    for j in range(X.shape[0] - 1):
        Q.append(f"{S1}{'              '}{j + 1}")
        Q.append(f"{S1}{'              '}{S1 * 2}")

    if lsprt is not None and lsprt != 0:
        Q.append(f"{'              '}L/S")
        Q.append(f"{S1}{'              '}{S1 * 2}")
    # else:
    #     Q.append(f"{S1}{'              '}{X.shape[0]}{S1 * 2}")
    #     Q.append(f"{S1}{'              '}{S1 * 2}")

    B = [f"{q}{b}" for q, b in zip(Q, B)]

    c = f"{S1 * 21}xret       alpha"
    # for head in heads:
    #     c += f"{' ' * (9 - len(head))}{head}"
    # c = f"{S1 * 21}xret"
    for idx, head in enumerate(heads):
        spaces = " " * (max_col_widths[idx] - len(head))
        c += f"{spaces}{head}"

    print("     ----------------------------------------------------------------------------------------------------- ")
    print(f"              {www}portfolio sort, {res['hperiod']}-month holding period")
    print("              Excess returns, alphas, and loadings on:")
    print(f"              {mdl}")
    print("     ----------------------------------------------------------------------------------------------------- ")
    print(c)
    print("\n".join(B))
    print("     ----------------------------------------------------------------------------------------------------- ")

    return


# call below to test
# results = runUnivSort(ret.to_numpy(), indFF, me.to_numpy(), dates.values.flatten(), factorModel=1, printResults=1, plotFigure=0)
# prt_sort_results(results, lsprt=1)
# prt_sort_results(results, lsprt=0)
# results2 = runUnivSort(ret.to_numpy(), indFF, me.to_numpy(), dates.values.flatten(), factorModel=6, printResults=1, plotFigure=0)
# prt_sort_results(results2, lsprt=1)


import numpy as np
import pandas as pd
from AssayingAnomalies.Functions.makeUnivSortInd import makeUnivSortInd


def makeBivSortInd(var1, ptfNumThresh1, var2, ptfNumThresh2, sort_type='unconditional', breaksFilterInd=None,
                   portfolioMassInd=None):
    """
    Creates a portfolio index matrix for a bivariate sort indicating which portfolio each stock-month belongs to.
    --------------------------------------------------------------------------------------------------------------------
    Required Inputs:
        var1 : numpy.ndarray or pd.DataFrame
            Matrix with first variable used for sorting.
        ptfNumThresh1 : int or numpy.ndarray
            Scalar or vector used to determine the number of portfolios in the first direction.
        var2 : numpy.ndarray or pd.DataFrame
            Matrix with second variable used for sorting.
        ptfNumThresh2 : int or numpy.ndarray
            Scalar or vector used to determine the number of portfolios in the second direction.
    --------------------------------------------------------------------------------------------------------------------
    Optional Inputs:
        sort_type : str, default 'unconditional'
            String indicating whether bivariate sort is 'unconditional' or 'conditional'.
        breaksFilterInd : numpy.ndarray or pd.DataFrame, default None
            Optional matrix used to filter portfolio breakpoints.
        portfolioMassInd : numpy.ndarray or pd.DataFrame, default None
            Optional matrix used for portfolio mass indicators.
    --------------------------------------------------------------------------------------------------------------------
    Returns:
        ind : numpy.ndarray
            A matrix that indicates the portfolio each stock-month falls under.
    --------------------------------------------------------------------------------------------------------------------
    Examples:
        # 5x5 unconditional sort on size and momentum
        ind = makeBivSortInd(me, 5, R, 5)

        # 2x3 (FF-style tertiles) unconditional sort on size and momentum
        ind = makeBivSortInd(me, 2, R, [30, 70])

        # 5x5 conditional sort on size and momentum
        ind = makeBivSortInd(me, 5, R, 5, sort_type='conditional')

        # 5x5 unconditional sort on size and momentum with breaksFilterInd
        ind = makeBivSortInd(me, 5, R, 5, breaksFilterInd=breaksFilterInd)

        # 5x5 unconditional sort on size and momentum with portfolioMassInd
        ind = makeBivSortInd(me, 5, R, 5, portfolioMassInd=portfolioMassInd)

        # 5x5 conditional sort on size and momentum with breaksFilterInd
        ind = makeBivSortInd(me, 5, R, 5, sort_type='conditional', breaksFilterInd=NYSE)
    --------------------------------------------------------------------------------------------------------------------
    """
    # Validate inputs
    if not (isinstance(var1, pd.DataFrame) or isinstance(var1, np.ndarray)):
        raise ValueError("Input 'var1' must be a DataFrame or a NumPy array.")

    if not (isinstance(var2, pd.DataFrame) or isinstance(var2, np.ndarray)):
        raise ValueError("Input 'var2' must be a DataFrame or a NumPy array.")

    # Convert input DataFrames to Numpy arrays
    if isinstance(var1, pd.DataFrame):
        var1 = var1.values

    if isinstance(var2, pd.DataFrame):
        var2 = var2.values

    if not np.isscalar(ptfNumThresh1) and not isinstance(ptfNumThresh1, (np.ndarray, list)):
        raise ValueError("Input 'ptfNumThresh1' must be a scalar, list, or a NumPy array.")

    if not np.isscalar(ptfNumThresh2) and not isinstance(ptfNumThresh2, (np.ndarray, list)):
        raise ValueError("Input 'ptfNumThresh2' must be a scalar, list, or a NumPy array.")

    if sort_type not in ['unconditional', 'conditional']:
        raise ValueError("sortType must be either 'unconditional' or 'conditional'.")

    if breaksFilterInd is not None:
        if not (isinstance(breaksFilterInd, pd.DataFrame) or isinstance(breaksFilterInd, np.ndarray)):
            raise ValueError("Input 'breaksFilterInd' must be a DataFrame, a NumPy array, or None.")
        if var1.shape != breaksFilterInd.shape or var2.shape != breaksFilterInd.shape:
            raise ValueError("Input 'breaksFilterInd' must have the same shape as 'var1 and 'var2'.")

    if portfolioMassInd is not None:
        if not (isinstance(portfolioMassInd, pd.DataFrame) or isinstance(portfolioMassInd, np.ndarray)):
            raise ValueError("Input 'portfolioMassInd' must be a DataFrame, a NumPy array, or None.")
        if var1.shape != portfolioMassInd.shape or var2.shape != portfolioMassInd.shape:
            raise ValueError("Input 'portfolioMassInd' must have the same shape as 'var1' and 'var2'.")

    # If user passed a list instead of numpy array for ptfNumThresh, we will need to change it to numpy array
    if isinstance(ptfNumThresh1, list):
        ptfNumThresh1 = np.array(ptfNumThresh1)
    if isinstance(ptfNumThresh2, list):
        ptfNumThresh2 = np.array(ptfNumThresh2)

    ind = np.zeros(var1.shape)

    # Sort based on the first variable
    ind1 = makeUnivSortInd(var1, ptfNumThresh1, breaksFilterInd=breaksFilterInd, portfolioMassInd=portfolioMassInd)

    # Store the number of portfolios
    n1 = int(np.nanmax(ind1))

    # Check the double sort type
    if sort_type == 'unconditional':
        # If unconditional
        # Sort based on the second variable
        ind2 = makeUnivSortInd(var2, ptfNumThresh2, breaksFilterInd=breaksFilterInd, portfolioMassInd=portfolioMassInd)

        # Store the number of portfolios
        n2 = int(np.nanmax(ind2))

        # Create the combined n1 x n2 portfolios
        for i in range(1, n1 + 1):
            for j in range(1, n2 + 1):
                ind[(ind1 == i) & (ind2 == j)] = (i - 1) * n2 + j  # Give it a number, in the end, all will be from 1 to n1 * n2
    else:
        # If conditional
        for i in range(1, n1 + 1):
            # Sort within each of the portfolios sorted based on the first variable
            temp = var2.copy()
            temp[ind1 != i] = np.nan
            temp_ind = makeUnivSortInd(temp, ptfNumThresh2, breaksFilterInd=breaksFilterInd,
                                       portfolioMassInd=portfolioMassInd)
            n2 = np.nanmax(temp_ind)
            ind[(ind1 == i) & (temp_ind > 0)] = temp_ind[(ind1 == i) & (temp_ind > 0)] + n2 * (i - 1)

    return ind


# Test the function
# import scipy.io
# import os
#
# path = r"C:\Users\josh\OneDrive - University at Buffalo\Desktop\Spring_2023\AssayingAnomalies-main\AssayingAnomalies-main\Data" + os.sep
# R = scipy.io.loadmat(path + 'R.mat')['R']
# # R = pd.read_csv(path + 'R.csv', index_col=0)
# me = scipy.io.loadmat(path + 'me.mat')['me']
# NYSE = scipy.io.loadmat(path + 'NYSE.mat')['NYSE']
# test = makeBivSortInd(me, 5, R, 5)
# test = makeBivSortInd(me, 2, R, [30, 70])
# test = makeBivSortInd(me, 5, R, 5, sort_type='conditional')



from AssayingAnomalies.wrds_utilities.utilities import *
from datetime import datetime


def remote_get_crsp_data(params, **kwargs):
    "Timekeeping"
    print(f"\nNow working on collecting data from CRSP. Run started at {datetime.now()}.\n")

    # Default parameter values
    p = {
        "delete_files_ok": True
    }

    # Update parameters with user-specified values (if provided).
    for key, value in kwargs.items():
        if key in p:
            p[key] = value

    # Define the start and end dates.
    start = params.sample_start
    end = params.sample_end

    """Want to download and save the following tables from CRSP:
     MSFHDR, MSF, MSEDELIST, MSEEXCHDATES, CCMXPF_LNKHIST, STOCKNAMES"""
    # Create commands for custom sql queries and saving files
    commands = []
    file_locations = []

    sql_string_msfhdr = "select * from CRSP.MSFHDR"
    file_name_msfhdr = "crsp_msfhdr.csv"
    file_location_msfhdr = f'/scratch/rochester/{file_name_msfhdr}'
    file_locations.append(file_location_msfhdr)

    sql_string_msf = f"select * from CRSP.MSF " \
                     f"where date>='01-01-{start}' " \
                     f"and date<'01-01-{end+1}' "
    file_name_msf = "crsp_msf.csv"
    file_location_msf = f'/scratch/rochester/{file_name_msf}'
    file_locations.append(file_location_msf)

    sql_string_msedelist = "select * from CRSP.MSEDELIST"
    file_name_msedelist = "crsp_msedelist.csv"
    file_location_msedelist = f'/scratch/rochester/{file_name_msedelist}'
    file_locations.append(file_location_msedelist)

    sql_string_exchdates = "select * from CRSP.MSEEXCHDATES"
    file_name_exchdates = "crsp_mseexchdates.csv"
    file_location_exchdates = f'/scratch/rochester/{file_name_exchdates}'
    file_locations.append(file_location_exchdates)

    sql_string_lnkhist = "select * from CRSP.CCMXPF_LNKHIST"
    file_name_lnkhist = "crsp_ccmxpf_lnkhist.csv"
    file_location_lnkhist = f'/scratch/rochester/{file_name_lnkhist}'
    file_locations.append(file_location_lnkhist)

    sql_string_stocknames = "select * from CRSP.STOCKNAMES"
    file_name_stocknames = "crsp_stocknames.csv"
    file_location_stocknames = f'/scratch/rochester/{file_name_stocknames}'
    file_locations.append(file_location_stocknames)

    command1 = f"testing = db.raw_sql(\"{sql_string_msfhdr}\")"
    command2 = f"testing.to_csv(\"/scratch/rochester/{file_name_msfhdr}\")"
    command3 = f"testing = db.raw_sql(\"{sql_string_msf}\")"
    command4 = f"testing.to_csv(\"/scratch/rochester/{file_name_msf}\")"
    command5 = f"testing = db.raw_sql(\"{sql_string_msedelist}\")"
    command6 = f"testing.to_csv(\"/scratch/rochester/{file_name_msedelist}\")"
    command7 = f"testing = db.raw_sql(\"{sql_string_exchdates}\")"
    command8 = f"testing.to_csv(\"/scratch/rochester/{file_name_exchdates}\")"
    command9 = f"testing = db.raw_sql(\"{sql_string_lnkhist}\")"
    command10 = f"testing.to_csv(\"/scratch/rochester/{file_name_lnkhist}\")"
    command11 = f"testing = db.raw_sql(\"{sql_string_stocknames}\")"
    command12 = f"testing.to_csv(\"/scratch/rochester/{file_name_stocknames}\")"

    commands.extend([command1, command2, command3, command4, command5, command6, command7, command8, command9,
                     command10, command11, command12])

    # Start a new child process, and connect to the SSH server
    child = ssh_login(params)

    # Start a qrsh session
    start_qrsh(params, child)

    # Start Python session
    start_python_wrds(child)
    # Execute python commands on wrds cloud
    execute_commands(child, commands)

    # Exit all sessions
    exit_sessions(child)

    # Transfer the files and return a list of files that did not transfer
    print("Using SFTP to transfer files.")
    save_folder = params.crspFolder
    problem_files = transfer_files(params, file_locations, save_folder)

    # Delete the files
    if p['delete_files_ok']:
        delete_files(params, file_locations, problem_files)

    # Print statement indicating if all files were completely transferred or not.
    if problem_files:
        print("The following files did not transfer: ")
        for file in problem_files:
            print(file)
    else:
        print("All files were successfully transferred and deleted from WRDS storage.")

    "Timekeeping"
    print(f"\nFinished collecting data from CRSP. Run ended at {datetime.now()}.\n")

    return

# from AssayingAnomalies import Config
# import AssayingAnomalies.Functions as aa
#
#
# "Create an instance of class 'Config' "
# params = Config()
#
# "Prompt the user to enter their parameters"
# params.prompt_user()
#
# "Create folders to store the downloaded data and created variables"
# params.make_folders()
# remote_get_crsp_data(params)


import numpy as np
import pandas as pd

def makePastPerformance(returns, lag_start, lag_end):
    """
    PURPOSE: This function calculates cumulative returns for momentum variables.
    --------------------------------------
    :param returns: pandas DataFrame or numpy array
            - dates x permno monthly returns
    :param lag_start: int
            - how many periods ago does the lag start (including this month) ; e.g. lag_start = 12
    :param lag_end:
            - how many periods ago does the lag end (excluding this month); e.g. lag_end = 1
    ---------------------------------------
    :return: returns: numpy array
            - an array with cumulative returns
    ----------------------------------------
    EXAMPLES:
            Classic Momentum: from 12 months ago to one month ago (NOT including last month)
                - R = makePastPerformance(returns=ret, lag_start=12, lag_end=1)
    """
    # first make sure returns is a numpy array
    if isinstance(returns, pd.DataFrame):
        returns = returns.values

    n_rows, n_cols = returns.shape
    R = np.full((n_rows, n_cols), np.nan)

    # check to see if the number of lags is 1
    if lag_start - lag_end == 1:
        # if the number of lags is one then check to see if the starting lag is 1 and if it is then set R = returns.
        if lag_start == 1:
            R = returns
        # if the starting lag isn't one but the number of lags is one then set then create lagged returns matrix and
        # set the first 'lag_start' number of rows to nan.
        else:
            R[lag_start - 1:, :] = returns[:-lag_start + 1, :]
    else:
        # Set the missing returns to 0
        is_nan_ind = np.isnan(returns)
        returns[is_nan_ind] = 0

        # Use log returns for quicker calculation.
        tret = np.log(1 + returns)
        RR = np.full((n_rows, n_cols), np.nan)

        if lag_end == 0:
            RR = tret
        else:
            RR[lag_end:, :] = tret[:-lag_end, :]
            # RR[:lag_end, :] = np.nan
        # print(f"RR:")
        # print(RR)
        # calculate the cumulative log returns
        for i in range(lag_end + 1, lag_start):
            RR[i:, :] += tret[:-i, :]
            RR[:i, :] = np.nan
            # print(f"Iteration {i}:")
            # print(RR)

        R = RR
        # set the first 'frm - 1' rows to nan
        R[:lag_start - 1, :] = np.nan
        # put the nans back that were originally there.
        R[is_nan_ind] = np.nan
        # take exponential to get back returns
        R = np.exp(R)

    return R


# use the below lines to filter out stocks that don't have sharecodes 10 or 11.
# import pandas as pd
# saveFolder = r'/home/jlaws13/PycharmProjects/AssayingAnomalies_root/Data/CRSP/'
# shrcd = pd.read_csv(saveFolder + 'shrcd.csv', index_col=0).fillna(0).replace([np.inf, -np.inf], 0).astype(int)
# shrcd = shrcd.astype(int)
# colsToKeep = np.where(((shrcd == 10) | (shrcd == 11).any()))[1]
# colsToKeep = np.unique(colsToKeep)
# ret_dom = ret.reindex(columns=ret.columns[colsToKeep]) # TODO:Note When using reindex, nan values are set to zero!?
# ret_dom = ret.iloc[:, colsToKeep]
# test = makePastPerformance(ret_dom, 12, 1)

# use the below returns for a quick way to test the output in python and matlab, respectively.
# returns_python = pd.DataFrame({
#     'A': [0.01, -0.02, 0.03, -0.04, 0.05, -0.06, 0.07, -0.08, 0.09, -0.10, 0.11, -0.12, 0.13, -0.14, 0.15],
#     'B': [0.02, 0.03, -0.04, 0.01, -0.06, 0.07, -0.08, 0.05, -0.10, 0.11, -0.12, 0.09, -0.14, 0.15, -0.16],
#     'C': [-0.03, 0.04, 0.01, 0.02, 0.07, -0.08, 0.05, 0.06, 0.11, -0.12, 0.09, 0.10, 0.15, -0.16, 0.13],
#     'D': [0.04, -0.01, 0.02, 0.03, -0.08, 0.05, 0.06, 0.07, -0.12, 0.09, 0.10, 0.11, -0.16, 0.13, 0.14],
# })

#  returns_matlab = [
#  0.01,  0.02, -0.03,  0.04;
# -0.02,  0.03,  0.04, -0.01;
#  0.03, -0.04,  0.01,  0.02;
# -0.04,  0.01,  0.02,  0.03;
#  0.05, -0.06,  0.07, -0.08;
# -0.06,  0.07, -0.08,  0.05;
#  0.07, -0.08,  0.05,  0.06;
# -0.08,  0.05,  0.06,  0.07;
#  0.09, -0.10,  0.11, -0.12;
# -0.10,  0.11, -0.12,  0.09;
#  0.11, -0.12,  0.09,  0.10;
# -0.12,  0.09,  0.10,  0.11;
#  0.13, -0.14,  0.15, -0.16;
# -0.14,  0.15, -0.16,  0.13;
#  0.15, -0.16,  0.13,  0.14;
# ]

# test = makePastPerformance(returns_python, 12, 1)




from datetime import datetime
import numpy as np
import pandas as pd
import os
import math
from scipy.stats import truncnorm, multivariate_normal
from scipy.linalg import sqrtm
from multiprocessing import Pool
from .remote_get_dsi import remote_get_dsi
from .Gibbs import *


def process_permno(permno, price, trade_direction, pm, group_mask, n_sweeps_test, reg_draw_test, varu_draw_test,
                   q_draw_test, varu_start_test, c_start_test, beta_start_test, nDrop):
    results_for_permno = []
    first_year = price[permno].dropna().index.year.unique()[0]
    last_year = price[permno].dropna().index.year.unique()[-1]

    for year in range(first_year, last_year + 1):
        try:
            mask_ffill_filtered = group_mask.loc[group_mask.index.year == year, permno]
            # print(permno, str(year))

            # Filter the data for the current permno and year
            p_data = price.loc[price.index.year == year, permno]
            pm_data = pm[pm.index.year == year]
            q_data = trade_direction.loc[trade_direction.index.year == year, permno]
            # print(len(p_data))
            if np.sum(~np.isnan(p_data.to_numpy())) >= 60:
                p_data_0 = p_data[mask_ffill_filtered == 0].to_numpy()
                p_data_1 = p_data[mask_ffill_filtered == 1].to_numpy()
                p_data_2 = p_data[mask_ffill_filtered == 2].to_numpy()
                pm_data_0 = pm_data[mask_ffill_filtered == 0].values.flatten()
                pm_data_1 = pm_data[mask_ffill_filtered == 1].values.flatten()
                pm_data_2 = pm_data[mask_ffill_filtered == 2].values.flatten()
                q_data_0 = q_data[mask_ffill_filtered == 0].to_numpy()
                q_data_1 = q_data[mask_ffill_filtered == 1].to_numpy()
                q_data_2 = q_data[mask_ffill_filtered == 2].to_numpy()

                group = 0
                if np.sum(~np.isnan(p_data_0)) >= 60:
                    # print(f"group {group}")
                    # Apply RollGibbsBeta function
                    parmOut_0 = roll_gibbs_beta(p_data_0, pm_data_0, q_data_0, n_sweeps_test, reg_draw_test,
                                                varu_draw_test, q_draw_test, varu_start_test, c_start_test,
                                                beta_start_test)
                    # Process the output from RollGibbsBeta
                    p2 = parmOut_0[nDrop:]  # Drop the initial iterations (burn-in)
                    # p2 = parmOut  # Don't drop the initial iterations (burn-in)
                    # avg_results = pd.DataFrame(p2.mean(axis=0)).T  # Compute the average
                    avg_results = p2.mean(axis=0)  # Compute the average
                    # df_cost.loc[(year, group), permno] = avg_results[0]
                    # df_beta.loc[(year, group), permno] = avg_results[1]
                    # df_varu.loc[(year, group), permno] = avg_results[2]
                    result = (year, group, permno, avg_results[0], avg_results[1], avg_results[2])

                else:
                    # df_cost.loc[(year, group), permno] = np.nan
                    # df_beta.loc[(year, group), permno] = np.nan
                    # df_varu.loc[(year, group), permno] = np.nan
                    result = (year, group, permno, np.nan, np.nan, np.nan)

                results_for_permno.append(result)
                # Apply RollGibbsBeta function
                group = 1
                if np.sum(~np.isnan(p_data_1)) >= 60:
                    # print(f"group {group}")
                    # Apply RollGibbsBeta function
                    parmOut_1 = roll_gibbs_beta(p_data_1, pm_data_1, q_data_1, n_sweeps_test, reg_draw_test,
                                                varu_draw_test,
                                                q_draw_test, varu_start_test, c_start_test, beta_start_test)
                    # Process the output from RollGibbsBeta
                    p2 = parmOut_1[nDrop:]  # Drop the initial iterations (burn-in)
                    # p2 = parmOut  # Don't drop the initial iterations (burn-in)
                    # avg_results = pd.DataFrame(p2.mean(axis=0)).T  # Compute the average
                    avg_results = p2.mean(axis=0)  # Compute the average
                    # df_cost.loc[(year, group), permno] = avg_results[0]
                    # df_beta.loc[(year, group), permno] = avg_results[1]
                    # df_varu.loc[(year, group), permno] = avg_results[2]
                    result = (year, group, permno, avg_results[0], avg_results[1], avg_results[2])

                else:
                    # df_cost.loc[(year, group), permno] = np.nan
                    # df_beta.loc[(year, group), permno] = np.nan
                    # df_varu.loc[(year, group), permno] = np.nan
                    result = (year, group, permno, np.nan, np.nan, np.nan)

                results_for_permno.append(result)
                # Apply RollGibbsBeta function
                group = 2
                if np.sum(~np.isnan(p_data_2)) >= 60:
                    # print(f"group {group}")
                    parmOut_2 = roll_gibbs_beta(p_data_2, pm_data_2, q_data_2, n_sweeps_test, reg_draw_test,
                                                varu_draw_test, q_draw_test, varu_start_test, c_start_test,
                                                beta_start_test)
                    p2 = parmOut_2[nDrop:]
                    avg_results = p2.mean(axis=0)  # Compute the average
                    # df_cost.loc[(year, group), permno] = avg_results[0]
                    # df_beta.loc[(year, group), permno] = avg_results[1]
                    # df_varu.loc[(year, group), permno] = avg_results[2]
                    result = (year, group, permno, avg_results[0], avg_results[1], avg_results[2])

                else:
                    # df_cost.loc[(year, group), permno] = np.nan
                    # df_beta.loc[(year, group), permno] = np.nan
                    # df_varu.loc[(year, group), permno] = np.nan
                    result = (year, group, permno, np.nan, np.nan, np.nan)

                results_for_permno.append(result)

            else:
                # print(f"Permno {permno} in year {year} has less than 60 observations.")
                continue

        except Exception as e:
            print(f"Error processing permno {permno} in year {year}: {e}")
            results_for_permno.append((year, np.nan, permno, np.nan, np.nan, np.nan))
            continue
    # print(f"Completed permno {permno} which is {price.columns.get_loc(permno) + 1}/{len(permno.columns)}")
    print(f"Completed permno {permno}")
    return results_for_permno


def parallel_processing(params, price, trade_direction, pm, group_mask, n_sweeps_test, reg_draw_test, varu_draw_test,
                        q_draw_test, varu_start_test, c_start_test, beta_start_test, nDrop):
    permnos = price.columns.tolist()
    all_results = []

    # Create a pool of workers
    with Pool(params.num_cpus) as p:
        all_results = p.starmap(process_permno, [(permno, price, trade_direction, pm, group_mask, n_sweeps_test,
                                                  reg_draw_test, varu_draw_test, q_draw_test, varu_start_test,
                                                  c_start_test, beta_start_test, nDrop) for permno in permnos])

    # Flatten the list of results
    flat_results = [item for sublist in all_results for item in sublist]

    return flat_results


def crspGibbsBuildv01(params):
    # Store the daily CRSP data path
    daily_crsp_path = params.daily_crsp_folder + os.sep
    crsp_path = params.crspFolder + os.sep
    gibbs_path = params.gibbs_data_folder + os.sep

    # Load exchange codes
    dexchcd = pd.read_csv(daily_crsp_path + 'dexchcd.csv', index_col=0)

    # Find when permnos changed exchanges
    exch_change = dexchcd - dexchcd.shift(1)
    exch_change_mask = exch_change.notna() & (exch_change != 0)

    # Find when permnos issued shares
    dcfacpr = pd.read_csv(daily_crsp_path + 'dcfacpr.csv', index_col=0).astype(float)
    dcfacpr_change = (dcfacpr - dcfacpr.shift(1)) / dcfacpr
    dcfacpr_change_mask = (np.abs(dcfacpr_change) >= 0.2)


    # Calculate trade direction, q = sign of price_1 - price_0.
    dret = pd.read_csv(daily_crsp_path + 'dret.csv', index_col=0).astype(float)
    dprc = pd.read_csv(daily_crsp_path + 'dprc.csv', index_col=0).astype(float)
    price = np.log(1 + dret).cumsum()
    dp = np.abs(dprc) - np.abs(dprc.shift(1))
    trade_direction = np.sign(dp)
    trade_direction[dprc < 0] = 0
    trade_direction[(dp.isna()) | (dp == 0)] = 1

    # Download the daily stock index file and import it
    # remote_get_dsi(params=params)
    crsp_dsi = pd.read_csv(daily_crsp_path + 'crsp_dsi.csv', index_col=0).astype(float)
    pm = np.log(1 + crsp_dsi).cumsum()

    # Mask to put each permno in groups 0-2
    mask1 = exch_change_mask * 1
    mask2 = dcfacpr_change_mask * 2
    mask = mask1.values + mask2
    mask.index = pd.to_datetime(mask.index, format='%Y%m%d')
    mask.columns = price.columns
    mask[mask >= 2] = 2

    def custom_ffill(group):
        # Replace 0 with NaN, but keep the first 0 in a sequence of 0s
        group = group.replace(to_replace=0, method='ffill', limit=1).replace(to_replace=0, value=np.nan)
        # Forward fill NaN values
        return group.ffill()

    # Group by year and apply the custom forward fill within each year.
    group_mask = mask.groupby(mask.index.year).apply(lambda x: custom_ffill(x))

    # Now each day is 0, 1, 2 depending on which group.
    # Replace remaining NaNs with 0 if needed
    group_mask.fillna(0, inplace=True)

    "Calculated rolling gibbs beta"
    # First change indices to datetime
    trade_direction.index = pd.to_datetime(trade_direction.index, format='%Y%m%d')
    price.index = pd.to_datetime(price.index, format='%Y%m%d')
    pm.index = trade_direction.index

    n_sweeps_test = 1000
    reg_draw_test = True
    varu_draw_test = True
    q_draw_test = True
    varu_start_test = 0
    c_start_test = 0
    beta_start_test = 1

    # Number of results to throw away
    nDrop = 200

    # Create the multi-index
    groups = [0, 1, 2]
    years = mask.index.year.unique()
    multi_index = pd.MultiIndex.from_product([years, groups], names=['year', 'group'])

    # Initialize the dataframes with the multi-index
    df_cost = pd.DataFrame(index=multi_index, columns=price.columns)
    df_beta = pd.DataFrame(index=multi_index, columns=price.columns)
    df_varu = pd.DataFrame(index=multi_index, columns=price.columns)

    if not params.remote_or_not:
        # Iterate over each permno and year
        for permno in dret.columns:
        # for permno in dret.columns[:5]:
            # print(permno)
            first_year = price[permno].dropna().index.year.unique()[0]
            last_year = price[permno].dropna().index.year.unique()[-1]
            for year in range(first_year, last_year+1):
                mask_ffill_filtered = group_mask.loc[group_mask.index.year == year, permno]
                print(permno, str(year))
                # Filter the data for the current permno and year
                p_data = price.loc[price.index.year == year, permno]
                pm_data = pm[pm.index.year == year]
                q_data = trade_direction.loc[trade_direction.index.year == year, permno]
                # print(len(p_data))
                if np.sum(~np.isnan(p_data.to_numpy())) >= 60:
                    p_data_0 = p_data[mask_ffill_filtered == 0].to_numpy()
                    p_data_1 = p_data[mask_ffill_filtered == 1].to_numpy()
                    p_data_2 = p_data[mask_ffill_filtered == 2].to_numpy()
                    pm_data_0 = pm_data[mask_ffill_filtered == 0].values.flatten()
                    pm_data_1 = pm_data[mask_ffill_filtered == 1].values.flatten()
                    pm_data_2 = pm_data[mask_ffill_filtered == 2].values.flatten()
                    q_data_0 = q_data[mask_ffill_filtered == 0].to_numpy()
                    q_data_1 = q_data[mask_ffill_filtered == 1].to_numpy()
                    q_data_2 = q_data[mask_ffill_filtered == 2].to_numpy()

                    group = 0
                    if np.sum(~np.isnan(p_data_0)) >= 60:
                        # print(f"group {group}")
                        # Apply RollGibbsBeta function
                        parmOut_0 = roll_gibbs_beta(p_data_0, pm_data_0, q_data_0, n_sweeps_test, reg_draw_test, varu_draw_test, q_draw_test, varu_start_test, c_start_test, beta_start_test)
                        # Process the output from RollGibbsBeta
                        p2 = parmOut_0[nDrop:]  # Drop the initial iterations (burn-in)
                        # p2 = parmOut  # Don't drop the initial iterations (burn-in)
                        # avg_results = pd.DataFrame(p2.mean(axis=0)).T  # Compute the average
                        avg_results = p2.mean(axis=0)  # Compute the average
                        df_cost.loc[(year, group), permno] = avg_results[0]
                        df_beta.loc[(year, group), permno] = avg_results[1]
                        df_varu.loc[(year, group), permno] = avg_results[2]
                    else:
                        df_cost.loc[(year, group), permno] = np.nan
                        df_beta.loc[(year, group), permno] = np.nan
                        df_varu.loc[(year, group), permno] = np.nan

                        # Apply RollGibbsBeta function
                    group = 1
                    if np.sum(~np.isnan(p_data_1)) >= 60:
                        # print(f"group {group}")
                        # Apply RollGibbsBeta function
                        parmOut_1 = roll_gibbs_beta(p_data_1, pm_data_1, q_data_1, n_sweeps_test, reg_draw_test, varu_draw_test,
                                                    q_draw_test, varu_start_test, c_start_test, beta_start_test)
                        # Process the output from RollGibbsBeta
                        p2 = parmOut_1[nDrop:]  # Drop the initial iterations (burn-in)
                        # p2 = parmOut  # Don't drop the initial iterations (burn-in)
                        # avg_results = pd.DataFrame(p2.mean(axis=0)).T  # Compute the average
                        avg_results = p2.mean(axis=0)  # Compute the average
                        df_cost.loc[(year, group), permno] = avg_results[0]
                        df_beta.loc[(year, group), permno] = avg_results[1]
                        df_varu.loc[(year, group), permno] = avg_results[2]
                    else:
                        df_cost.loc[(year, group), permno] = np.nan
                        df_beta.loc[(year, group), permno] = np.nan
                        df_varu.loc[(year, group), permno] = np.nan

                        # Apply RollGibbsBeta function
                    group = 2
                    if np.sum(~np.isnan(p_data_2)) >= 60:
                        # print(f"group {group}")
                        parmOut_2 = roll_gibbs_beta(p_data_2, pm_data_2, q_data_2, n_sweeps_test, reg_draw_test, varu_draw_test, q_draw_test, varu_start_test, c_start_test, beta_start_test)
                        p2 = parmOut_2[nDrop:]
                        avg_results = p2.mean(axis=0)  # Compute the average
                        df_cost.loc[(year, group), permno] = avg_results[0]
                        df_beta.loc[(year, group), permno] = avg_results[1]
                        df_varu.loc[(year, group), permno] = avg_results[2]
                    else:
                        df_cost.loc[(year, group), permno] = np.nan
                        df_beta.loc[(year, group), permno] = np.nan
                        df_varu.loc[(year, group), permno] = np.nan

        # Save Dataframes
        df_cost.to_csv(gibbs_path + 'gibbs_cost.csv')
        df_beta.to_csv(gibbs_path + 'gibbs_beta.csv')
        df_varu.to_csv(gibbs_path + 'gibbs_varu.csv')

        print(f"Completed Hasbrouck's (2009) Gibbs construction at {datetime.now()}")
        return

    else:
        # Call the parallel process function
        results = parallel_processing(params, price, trade_direction, pm, group_mask, n_sweeps_test, reg_draw_test, varu_draw_test, q_draw_test, varu_start_test, c_start_test, beta_start_test, nDrop)
        print(f"Completed Hasbrouck's (2009) Gibbs construction at {datetime.now()}")
        for i, (year, group, permno, cost, beta, varu) in enumerate(results):
            if isinstance(group, float) and math.isnan(group):
                # print(f"NaN found in group at index {i}: (year: {year}, group: {group}, permno: {permno}, cost: {cost}, beta: {beta}, varu: {varu})")
                continue
            else:
                df_cost.loc[(year, group), permno] = cost
                df_beta.loc[(year, group), permno] = beta
                df_varu.loc[(year, group), permno] = varu
        df_cost.to_csv(gibbs_path + 'gibbs_cost.csv')
        df_beta.to_csv(gibbs_path + 'gibbs_beta.csv')
        df_varu.to_csv(gibbs_path + 'gibbs_varu.csv')

        return


# "To test against sample data:"
# df_cost = pd.read_csv(daily_crsp_path + 'gibbs_cost.csv', index_col=0).astype(float)
# df_beta = pd.read_csv(daily_crsp_path + 'gibbs_beta.csv', index_col=0).astype(float)
# df_varu = pd.read_csv(daily_crsp_path + 'gibbs_varu.csv', index_col=0).astype(float)
#
# # Create Gibbs
# # Reset the index to turn the multi-index into columns. Group by 'year' and calculate the mean for each 'permno'
# gibbs_cost = df_cost.astype(float).groupby(level=0).mean()
# gibbs_beta = df_beta.astype(float).groupby(level=0).mean()
# gibbs_varu = df_varu.astype(float).groupby(level=0).mean()
#
#
# # Load test data
# crspgibbs = pd.read_csv('/scratch/jlaws13/AA_Data/Gibbs/crspgibbs.csv', low_memory=False)
# # Keeping the first occurrence
# no_duplicates = crspgibbs.groupby(['year', 'permno']).mean().reset_index()
# test_cost = pd.pivot(no_duplicates, index='year', columns='permno', values='c')
# test_beta = pd.pivot(no_duplicates, index='year', columns='permno', values='beta')
# test_varu = pd.pivot(no_duplicates, index='year', columns='permno', values='varu')
#
#
# # Aligning with my dataframe
# test_cost = test_cost.loc[params.sample_start:params.sample_end]
# test_beta = test_beta.loc[params.sample_start:params.sample_end]
# test_varu = test_varu.loc[params.sample_start:params.sample_end]
# test_cost.columns = test_cost.columns.astype(str)
# test_beta.columns = test_beta.columns.astype(str)
# test_varu.columns = test_varu.columns.astype(str)
# # This adds any missing columns from df_cost into test_cost with NaN values
# test_cost = test_cost.reindex(columns=df_cost.columns, fill_value=np.nan)
# test_beta = test_beta.reindex(columns=df_cost.columns, fill_value=np.nan)
# test_varu = test_varu.reindex(columns=df_cost.columns, fill_value=np.nan)
#
#
# correlation_cost = {}
# correlation_beta = {}
# correlation_varu = {}
# for col in df_cost.columns:
#     if len(gibbs_beta[col].dropna()) > 4:
#         correlation_cost[col] = gibbs_cost[col].dropna().corr(test_cost[col].dropna())
#         correlation_beta[col] = gibbs_beta[col].dropna().corr(test_beta[col].dropna())
#         correlation_varu[col] = gibbs_varu[col].dropna().corr(test_varu[col].dropna())
#     else:
#         correlation_cost[col] = np.nan
#         correlation_beta[col] = np.nan
#         correlation_varu[col] = np.nan
#
# correlation_cost = pd.DataFrame(list(correlation_cost.items()), columns=['permno', 'correlation'])
# correlation_beta = pd.DataFrame(list(correlation_beta.items()), columns=['permno', 'correlation'])
# correlation_varu = pd.DataFrame(list(correlation_varu.items()), columns=['permno', 'correlation'])
# print(f"Cost correlation = {correlation_cost['correlation'].mean()}")
# print(f"Beta correlation = {correlation_beta['correlation'].mean()}")
# print(f"Varu correlation = {correlation_varu['correlation'].mean()}")
#
#



import numpy as np
import pandas as pd
from datetime import datetime
import os
from AssayingAnomalies.Functions.runUnivSort import runUnivSort
from AssayingAnomalies.Functions.makeUnivSortInd import makeUnivSortInd
from AssayingAnomalies.Functions.rank_with_nan import rank_with_nan


def makeUniverses(params):
    """This function creates and stores a structure that contains indicators for small/large caps based on two
    classifications: Fama-French (< or > than NYSE 50th percentile) and Russell ((not) in the top 1000 stocks based on
    market cap).
    ------------------------------------------------------------------------------------------
    Required Inputs:

    ------------------------------------------------------------------------------------------
    Output:
        -None

    ------------------------------------------------------------------------------------------
    Examples:
    makeUniverses(Params)
    ------------------------------------------------------------------------------------------
    Dependencies:
        -Uses rowrank(), runUnivSort()
    ------------------------------------------------------------------------------------------
    References
    1. Novy-Marx, R. and M. Velikov, 2023, Assaying anomalies, Working paper.

    """

    # Timekeeping
    print(f"\n\n\nNow creating universes. Run began at {datetime.now()}.\n")

    # Set dataPaths
    # crspFolder = r'/home/jlaws13/PycharmProjects/AssayingAnomalies_root/Data/CRSP/'
    crspFolder = params.crspFolder + os.sep

    ret = pd.read_csv(crspFolder + 'ret.csv', index_col=0)
    me = pd.read_csv(crspFolder + 'me.csv', index_col=0)
    me.index = me.index.astype(int)
    me.columns = pd.to_numeric(me.columns, errors='coerce').astype(int)
    NYSE = pd.read_csv(crspFolder + 'NYSE.csv', index_col=0)
    ff = pd.read_csv(crspFolder + 'ff.csv', index_col=0)
    dates = pd.read_csv(crspFolder + 'dates.csv', index_col=0)

    # create a matrix ranking each stock from largest to smallest each year
    # rme = pd.DataFrame(rankdata(-me, method='average', axis=1, nan_policy='omit')) # TODO:Note Once I update to scipy >= 1.10.0, this line will replace the following 9 lines of code.
    # rme = np.zeros_like(me)
    # for i in range(len(me)):
    #     temp = np.array(me.iloc[i, :]) # select ith row of me
    #     new_ar = np.nan * np.zeros_like(temp) # create array of nan's same length as 'temp'
    #     mask_indx = np.where(~np.isnan(temp))[0] # location of non-nan values in 'temp'
    #     temp = temp[~np.isnan(temp)] # removes nan values
    #     temp = rankdata(-temp) # ranks the stocks from largest to smallest
    #     new_ar[mask_indx] = temp # replaces the nan values in new_ar with corresponding ranks
    #     rme[i] = new_ar # sets ith row of rme.
    rme = pd.DataFrame(rank_with_nan(-me))
    rme.columns = me.columns
    rme.index = me.index
    rme.columns = rme.columns.astype(int)
    # rme.columns = pd.to_numeric(rme.columns, errors='coerce').astype(int)
    # checking apples rank in 202012. Should return 1
    if dates.values.flatten()[-1] >=202012:
        print("Apple's rank as of December 2020 is " + str(rme.loc[202012, 14593]))


    "The next section of code relies on a function 'makeUnivSortInd(me, 2, NYSE) which I will create below and then move to a seperate file later on"
    # Create empty dictionary to store FF and Russell results
    universes = {}

    # Do the FF universe first
    indFF = makeUnivSortInd(me, 2, NYSE)  # make FF index

    # Get factor loadings, portfolio returns, etc.
    results_ff = runUnivSort(params=params, ret=ret, ind=indFF, mcap=me, dates=dates, factorModel=1, printResults=1, plotFigure=0)

    # Create dictionary that will be used to store FF universe data
    universe_1 = {
        'head': 'ff',
        'ind': indFF,
        'res': results_ff
    }

    # create excess portfolio returns and store in dictionary
    universe_1['xret'] = universe_1['res']['pret'][:, :2] - np.column_stack((ff.rf.values, ff.rf.values))

    # And the Russell universe next
    indRuss = np.zeros_like(rme)
    indRuss[rme > 1000] = 1
    indRuss[rme <= 1000] = 2
    indRuss[indRuss == 0] = np.nan

    results_russell = runUnivSort(params=params, ret=ret, ind=indRuss, mcap=me, dates=dates, factorModel=1, printResults=1, plotFigure=0)

    # Create dictionary that will be used to store FF universe data
    universe_2 = {
        'head': 'Russell',
        'ind': indRuss,
        'res': results_russell
    }

    # create excess portfolio returns and store in dictionary
    universe_2['xret'] = universe_2['res']['pret'][:, :2] - np.column_stack((ff.rf.values, ff.rf.values))

    # store both universes
    universes[universe_1['head']] = universe_1
    universes[universe_2['head']] = universe_2

    # saving the universes dictionary as a pickle for now # :TODO:N
    pd.to_pickle(universes, crspFolder + 'universe.pkl')

    # Make cumulative market cap percentile. Add a tiny bit of noise to ensure proper bucketing
    tempme = me + (np.random.rand(*me.shape) - 0.5) / 1_000_000
    rme = rank_with_nan(-tempme)

    # Initialize the cumulative market cap percentile matrix
    mep = np.empty(me.shape)
    mep[:] = np.nan

    # Store number of months
    n_months = me.shape[0]

    # Change 'me' to numpy for the loop below to work
    me = me.to_numpy()

    # loop over the months
    for i in range(n_months):
        temp = me[i, :]
        ii = np.isfinite(temp)
        temp = temp[ii]
        temp = np.cumsum(-np.sort(-temp))
        temp = temp / temp[-1]
        jj = rme[i, ii].astype(int) - 1
        mep[i, ii] = temp[jj]

    # Save the market cap percentile matrix.
    mep = pd.DataFrame(mep, columns=NYSE.columns, index=NYSE.index)
    mep.to_csv(crspFolder + 'mep.csv')

    # Timekeeping
    print(f"\n\n\nUniverse creation completed at {datetime.now()}.\n")

    return

# makeUniverses()



from AssayingAnomalies.wrds_utilities.utilities import ssh_login, start_qrsh, start_python_wrds, exit_sessions,\
    execute_commands, transfer_files, delete_files, create_virtual_environment
from datetime import datetime
import numpy as np


def remote_get_daily_crsp(params, **kwargs):
    # Create a virtual environment for pyarrow and parquet
    print(f"First we need to create a virtual environment.")
    create_virtual_environment(params)

    # Timekeeping
    print(f"Now downloading daily CRSP stock file. Run began at {datetime.now()}.")

    # Default keyword arguments
    p = {
        "delete_files_ok": True
    }

    # Update keyword arguments with user-specified values (if provided).
    for key, value in kwargs.items():
        if key in p:
            p[key] = value

    # Define the columns to pull from WRDS
    columns_to_pull = ['permno', 'date', 'cfacpr', 'cfacshr', 'bidlo', 'askhi', 'prc', 'vol', 'ret', 'bid',
                       'ask', 'shrout', 'openprc', 'numtrd']

    # Set the beginning and ending year
    start = params.sample_start
    end = params.sample_end

    # Manually select years before 2000 based on your intervals
    years_before_2000 = [start, 2000 - 55, 2000 - 25, 2000 - 10]  # 1925, 1970, 1985, 1990

    # Ensure years_before_2000 are within the range and unique
    years_before_2000 = [year for year in years_before_2000 if start <= year < 2000]
    years_before_2000 = sorted(set(years_before_2000))

    # Years from 2000 to end, split into three periods
    if end >= 2000:
        years_2000_to_end = np.linspace(max(start, 2000), end, 3, endpoint=True).astype(int)
    else:
        years_2000_to_end = []

    # Combine and sort the lists
    years = sorted(set(years_before_2000 + years_2000_to_end.tolist()))

    # Create commands for custom sql query and saving files
    commands = []
    file_locations = []
    for start, end in zip(years, years[1:]):
        # dsf_file_name = f"crsp_dsf_{end}.csv"
        dsf_file_name = f"crsp_dsf_{end}.parquet"
        file_location_dsf = f'/scratch/rochester/{dsf_file_name}'
        file_locations.append(file_location_dsf)
        dsf_sql_string = "SELECT " + ", ".join(columns_to_pull) + f" FROM CRSP.DSF " \
                f"WHERE date>='01-01-{start}' " \
                f"and date<'01-01-{end+1}' "
        command1 = f"testing = db.raw_sql(\"{dsf_sql_string}\")"
        command2 = f"testing.to_parquet(\"/scratch/rochester/{dsf_file_name}\")"
        commands.extend([command1, command2])

    # Create the delist file names, location, and commands
    delist_file_name = "crsp_dsedelist.csv"
    file_location_delist = f'/scratch/rochester/{delist_file_name}'
    file_locations.append(file_location_delist)
    delist_sql_string = "SELECT * FROM CRSP.DSEDELIST"
    command3 = f"testing = db.raw_sql(\"{delist_sql_string}\")"
    command4 = f"testing.to_csv(\"/scratch/rochester/{delist_file_name}\")"
    commands.extend([command3, command4])

    # Start a new child process, and connect to the SSH server
    child = ssh_login(params)

    # Start a qrsh session
    start_qrsh(params, child)

    # Start Python session
    start_python_wrds(child, venv=True)

    # Execute python commands on wrds cloud
    execute_commands(child, commands)

    # Exit all sessions
    exit_sessions(child)

    # Transfer the files and return a list of files that did not transfer
    print("Using SFTP to transfer files.")
    save_folder = params.daily_crsp_folder
    problem_files = transfer_files(params, file_locations, save_folder)

    # Delete the files
    if p['delete_files_ok']:
        delete_files(params, file_locations, problem_files)

    # Print statement indicating if all files were completely transferred or not.
    if problem_files:
        print("The following files did not transfer: ")
        for file in problem_files:
            print(file)
    else:
        print("All files were successfully transferred and deleted from WRDS storage.")

    return

# from AssayingAnomalies import Config
# import AssayingAnomalies.Functions as AA
#
#
# "Create an instance of class 'Config' "
# params = Config()
#
# "Prompt the user to enter their parameters"
# params.prompt_user()
#
# "Create folders to store the downloaded data and created variables"
# params.make_folders()
#
# get_daily_crsp(params)



import numpy as np
import pandas as pd


def calcPtfRets(ret, ind, mcap, hper, weighting):
    """
    Purpose: Calculates portfolio returns, number of stocks, and market capitalizations for each portfolio indicated
            by 'ind'
    -----------------------------------------------------------------------------------------------------------------------
    Usage:
    -----------------------------------------------------------------------------------------------------------------------
    Inputs:
        -ret - a matrix of stock returns
        -ind - a matrix of portfolio index
        -mcap - a matrix of market capitalization numbers
        -hper - ascalar indicating the ptf holding period
        weighting - weighting scheme (one of {'V', 'v', 'E', 'e'})
    Outputs:
        -pret - matrix of raw value-weighted portfolio returns
        -ptfNumStocks - matrix of number of firms in each portfolio
        -ptfMarketCap - matrix of total market capitalization of each portfolio
    -----------------------------------------------------------------------------------------------------------------------
    Examples:
    ptfRet, ptfNumStocks, ptfMarketCap = calcPtfRets(ret, ind, me, 1, 'e') # Equal-weighted returns with 1-month holding
                                                                            period.
    -----------------------------------------------------------------------------------------------------------------------
    Dependencies:
        -makeUniverses()
    -----------------------------------------------------------------------------------------------------------------------
    References:
        1. Novy-Marx, R. and M. Velikov, 2023, Assaying anomalies, Working paper.
    """

    # if isinstance(mcap, pd.DataFrame):
    #     mcap = mcap.to_numpy()
    #
    # if isinstance(ind, pd.DataFrame):
    #     ind = ind.to_numpy()

    # create lagged dataframes
    lme = np.roll(mcap, 1, axis=0)
    lind = np.roll(ind, 1, axis=0)
    # the code above takes whatever was in the last row and moves it to the first, which we must now correct.
    lme[0] = np.nan
    lind[0] = 0

    # trying it with numpy slicing
    # lme = mcap[1:, :]
    # lind = ind[1:, :]

    # Check weighting scheme
    if weighting == 'e' or weighting=='E':
        weightingMcap = np.ones_like(mcap)
    else:
        weightingMcap = lme

    # store a couple of variables
    nPtfs = np.max(np.max(ind)).astype(int)
    nMonths, nStocks = ret.shape

    def mode(arr):
        # count the frequency of each unique value in the array
        values, counts = np.unique(arr, return_counts=True)
        # locate the index of the most frequent value(s)
        index = np.nanargmax(counts)
        # return the most frequent value(s)
        return values[index] if counts[index] > 1 else None

    # Carry over the index in case we are not rebalancing every month
    nNonZero = np.sum(lind > 0, axis=1) # counts the number of stocks that are in at least one portfolio for each month
    indReb = np.where(nNonZero)[0] # find the column indices of the
    rebFreq = mode(indReb - np.roll(indReb, 1, axis=0))
    startMonth = np.where(nNonZero > 0)[0][0] + 1
    endMonth = min(np.where(nNonZero > 0)[0][-1] + rebFreq, nMonths)

    for i in range(startMonth, endMonth):
        if nNonZero[i] == 0:
            lind[i, :] = lind[i - 1, :]

    # Initialize the output matrices
    ptfNumStocks = np.zeros((nMonths, nPtfs))
    ptfMarketCap = np.zeros((nMonths, nPtfs))
    ptfRet = np.zeros((nMonths, nPtfs))

    # Loop over the portfolios
    for i in range(nPtfs):
        ptfInd = (lind == i+1).astype(float)
        # ptfInd = (lind == 1)

        if hper > 1:
            for h in range(1, hper):
                ptfInd = ptfInd | (lind[max(0, i-h), :] == i)

        # Make sure we are only using stock-months with available lagged market cap & return observations
        # ptfInd = 1*(ptfInd & np.isfinite(lme) & np.isfinite(ret))
        ptfInd = ptfInd.astype(bool) & np.isfinite(lme) & np.isfinite(ret)
        ptfInd = ptfInd.astype(float)
        ptfInd[ptfInd == 0] = np.nan

        # Calculate the number of stocks in the portfolio in each month
        # print(np.nansum(ptfInd, axis=1))
        ptfNumStocks[:, i] = np.nansum(ptfInd, axis=1)

        # Calculate the portfolios market cap in each month and then
        sumWghtMcap = np.nansum(weightingMcap*ptfInd, axis=1)
        ptfMarketCap[:, i] = sumWghtMcap # stores the portfolios market cap for each month
        sumWghtMcap = np.tile(sumWghtMcap, (nStocks, 1)).T # repeats the above column to form an dates x nStocks array

        # Calculate and stores the portfolio return
        ptfRet[:, i] = np.nansum(ptfInd*ret*weightingMcap / sumWghtMcap, axis=1)

    return ptfRet, ptfNumStocks, ptfMarketCap

#
# "Below is for testing the function"
# from AssayingAnomalies import Config
# import AssayingAnomalies.Functions as aa
# import os
#
# "Create an instance of class 'Config' "
# params = Config()
#
# "Prompt the user to enter their parameters"
# params.prompt_user()
#
# "Create folders to store the downloaded data and created variables"
# params.make_folders()
#
# crspFolder = params.crspFolder + os.sep
# NYSE = pd.read_csv(crspFolder + 'NYSE.csv', index_col=0)
# ret = pd.read_csv(crspFolder + 'ret.csv', index_col=0)
# me = pd.read_csv(crspFolder + 'me.csv', index_col=0)
# shrcd = pd.read_csv(crspFolder + 'shrcd.csv', index_col=0).fillna(0).replace([np.inf, -np.inf], 0).astype(int)
# shrcd = shrcd.astype(int)
# colsToKeep = np.where(((shrcd == 10) | (shrcd == 11).any()))[1]
# colsToKeep = np.unique(colsToKeep)
# ret = ret.iloc[:, colsToKeep]
# me = me.iloc[:, colsToKeep]
# NYSE = NYSE.iloc[:, colsToKeep]

# indFF = aa.makeUnivSortInd(-me, 2, NYSE) # the ff universe
# ptfRet, ptfNumStocks, ptfMarketCap = calcPtfRets(ret, indFF, me, 1, 'v')
# ptfRet, ptfNumStocks, ptfMarketCap = calcPtfRets(ret, indFF, me, 1, 'e')
# ptfRet, ptfNumStocks, ptfMarketCap = calcPtfRets(ret, indFF, me, 2, 'e') # :TODO:F I suspect that something with the holding period is still off
# indFF = makeUnivSortInd(-me, 10, NYSE)
# ptfRet, ptfNumStocks, ptfMarketCap = calcPtfRets(ret, indFF, me, 1, 'v')



"""
Things to note:
            - The script creates a subdirectory titled 'AA_Data' to store all the created variables.
            - Sometimes the script asks the user to enter their WRDS username and password twice.
            - Sometimes a 'FutureWarning' is returned while making CRSP monthly data, just ignore this, it doesn't hurt
            anything and will be fixed later.
            - For now, make sure sample dates includes 2012.
            - In the matlab version, data is stored in the same directory as the code. In this version, the user can
            specify where they want the data to live, since they might be using cloud storage/computing.
"""
from AssayingAnomalies import Config
import AssayingAnomalies.Functions as aa


def download_and_process_data(params):
    "Download & store all the CRSP data we'll need"
    if params.remote_or_not:
        aa.remote_get_crsp_data(params=params)
    else:
        aa.getCRSPData(params=params)

    "Make CRSP data"
    aa.makeCRSPMonthlyData(params=params)

    "Make additional CRSP variables"
    aa.makeCRSPDerivedVariables(params=params)

    "Download & store all the COMPUSTAT data we'll need (annual and quarterly)"
    if params.remote_or_not:
        aa.remote_get_compustat_data(params=params)
    else:
        aa.getCOMPUSTATData(params=params)

    "Merge CRSP and COMPUSTAT, store all variables"
    aa.mergeCRSPCOMP(params=params)

    "Make additional COMPUSTAT variables"
    aa.makeCOMPUSTATDerivedVariables(params=params)

    "Download & store all the daily CRSP data we'll need"
    if params.remote_or_not:
        aa.remote_get_daily_crsp(params=params)
    else:
        aa.get_crsp_daily_data(params=params)

    "Make the daily variables"
    aa.makeCRSPDailyData(params=params)

    "Make the daily derived variablres"
    aa.makeCRSPDailyDerivedVariables(params=params)

    "Make transaction costs"
    aa.makeTradingCosts(params=params)

    "Make anomalies"
    aa.makeNovyMarxVelikovAnomalies(params=params)

    "Make beta"
    aa.makeBetas(params=params)

    print("Data download and processing is now complete. Check out the file, use_library.py for examples of how to use "
          "the toolkit.")


def initial_setup():
    print("Welcome to the AssayingAnomalies toolkit. Let's get you started.")
    params = Config()
    params.set_up()
    params.save_params()

    # Ask the user if they want to proceed with data download and processing
    proceed = input("Configuration saved. Would you like to proceed with downloading data and making variables? (yes/no): ").strip().lower()
    if proceed == 'yes':
        download_and_process_data(params)
    else:
        print("Setup complete. You can run the data download and processing later.")


if __name__ == "__main__":
    initial_setup()





"""
This file contains a tutorial on how to implement various basic asset pricing techniques using the Toolkit. These
include univariate sorts, bivariate sorts, Fama-MacBeth regressions, and accounting for transaction costs.
"""

import pandas as pd
import os
from AssayingAnomalies.config import Config


# First load in your parameters
params = Config.load_params()

ret = pd.read_csv(params.crspFolder + os.sep + 'ret.csv', index_col=0).astype(float)
me = pd.read_csv(params.crspFolder + os.sep + 'me.csv', index_col=0).astype(float)
BM = pd.read_csv(params.compFolder + os.sep + 'BM.csv', index_col=0).astype(float)
dates = pd.read_csv(params.crspFolder + os.sep + 'dates.csv', index_col=0).astype(float)
NYSE = pd.read_csv(params.crspFolder + os.sep + 'NYSE.csv', index_col=0).astype(float)
ff = pd.read_csv(params.crspFolder + os.sep + 'ff.csv', index_col=0).astype(float)



import numpy as np
import pandas as pd
import os
import statsmodels.api as sm
from scipy.stats.mstats import winsorize
from .prt_FMB_results import prt_FMB_results


def run_Fama_MacBeth(y, RHS_variables, dates, **kwargs):
    """
    Perform a Fama-MacBeth regression using time series data. This function
    handles data preparation including lagging, winsorizing, and weighted regression,
    followed by the two-step Fama-MacBeth regression procedure.

    Parameters:
    y (numpy.ndarray or pd.DataFrame): The dependent variable in the regression,
                                       typically returns of assets, shaped as (time, firms).
    RHS_variables (list): A list of numpy.ndarrays or pd.DataFrames representing
                          independent variables. Each item should be of the shape (time, firms).
    dates (numpy.ndarray or pd.DataFrame): An array of dates corresponding to the time dimension.
    **kwargs: Keyword arguments for additional parameters:
        - numLags (int): Number of periods to lag variables. Default is 1.
        - timePeriod (list or int): Specified time period for the analysis. Default is full sample.
        - minobs (int): Minimum number of observations required. Default is 100.
        - weightMatrix (numpy.ndarray or pd.DataFrame): Matrix for weighted regression. Default is None.
        - trimIndicator (int): Indicator for whether to trim (1) or not (0). Default is 0.
        - winsorTrimPctg (float): Percentage for winsorizing. Default is 0 (no winsorizing).
        - printResults (int): Indicator for printing results (1) or not (0). Default is 1.
        - neweyWestLags (int): Number of lags for Newey-West standard errors. Default is 0 (OLS errors).
        - noConst (int): Include a constant (0) or not (1). Default is 0.
        - keepWarnings (int): Keep warnings (1) or not (0). Default is 0.

    Returns:
    dict: A dictionary containing:
        - 'beta': Coefficients from the cross-sectional regressions, [const, char1, char2, ... , charn]
        - 'bbhat': Averaged coefficients over time.
        - 't': T-statistics for the averaged coefficients.
        - 'Rbar2': Adjusted R-squared from each time period's regression.
        - 'mean_R2': Arithmetic mean of adjusted R-squared values.

    Raises:
    Exception: Outputs an error message and assigns NaN for regression failures.

    Notes:
    """
    # Convert input DataFrames to Numpy array
    if isinstance(y, pd.DataFrame):
        y = y.to_numpy()

    # Convert input DataFrames to Numpy arrays
    for i in range(len(RHS_variables)):
        if isinstance(RHS_variables[i], pd.DataFrame):
            RHS_variables[i] = RHS_variables[i].to_numpy()

    if isinstance(dates, pd.DataFrame):
        dates = dates.values.flatten().astype(int)

    # Default parameter values for **kwargs
    p = {
        'numLags': 1,
        'timePeriod': [dates[0], dates[-1]],
        'minobs': 100,
        'weightMatrix': None,
        'trimIndicator': 0,
        'winsorTrimPctg': 0,
        'printResults': 1,
        'neweyWestLags': 0,
        'noConst': 0,
        'keepWarnings': 0,
    }

    # Update parameters with user-specified values (if provided).
    for key, value in kwargs.items():
        if key in p:
            p[key] = value

    # Check if user entered a subsample
    if p['timePeriod'] != [dates[0], dates[-1]]:
        # Find the start date
        s = next(i for i, date in enumerate(dates) if date >= p['timePeriod'][0])

        # Find the end date, by finding the index 'e' of the last date in the 'dates' array that is <= specified end
        # date in 'p['timePeriod']'. "next()" retrieves first item produced by the generator expression
        if len(p['timePeriod']) == 2:
            # Whatever the user chose
            e = next(i for i, date in reversed(list(enumerate(dates))) if date <= p['timePeriod'][1])
        else:
            # Or the last date in the dates array
            e = len(dates) - 1

        # Subset the input arrays
        y = y[s:e + 1, :]
        for i in range(len(RHS_variables)):
            RHS_variables[i] = RHS_variables[i][s:e + 1, :]
        dates = dates[s:e + 1]

    # Store some constants
    T = len(dates)
    n_characteristics = len(RHS_variables)

    # Check if weighted least squares
    if p['weightMatrix']:
        if isinstance(p['weightMatrix'], pd.DataFrame):
            p['weightMatrix'] = p['weightMatrix'].to_numpy()
        # Lag the weight matrix
        p['weightMatrix'] = np.roll(p['weightMatrix'], shift=p['numLags'], axis=0)[p['numLags']:, :]

    # Lag the RHS variables
    if p['numLags'] > 0:
        y = y[:-p['numLags'], :]
        for i in range(n_characteristics):
            RHS_variables[i] = np.roll(RHS_variables[i], shift=p['numLags'], axis=0)[p['numLags']:, :]

    # Winsorize if necessary
    if p['winsorTrimPctg'] > 0:
        upper = 1 - p['winsorTrimPctg'] / 100
        lower = p['winsorTrimPctg'] / 100
        for i in range(n_characteristics):
            RHS_variables[i] = winsorize(RHS_variables[i], limits=[lower, upper], axis=1, nan_policy='omit')

    # Initialize lists for storage.
    coefficients = np.empty((T-1, n_characteristics + 1))
    adjusted_r_squared = np.empty(T-1)

    # First stage of FMB regression
    for t in range(T-1):
        X_t = [RHS_variables[j][t, :].flatten() for j in range(n_characteristics)]
        X_t = np.column_stack(X_t)
        X_t = sm.add_constant(X_t)

        Y_t = y[t, :]
        Y_t = Y_t.flatten()

        try:
            if p['weightMatrix']:
                W_t = p['weightMatrix'][t, :].flatten()
                results = sm.WLS(Y_t, X_t, weights=W_t, missing='drop', hasconst=True).fit()
            else:
                results = sm.OLS(Y_t, X_t, missing='drop', hasconst=True).fit()
            results.summary()
            coefficients[t, :] = results.params
            adjusted_r_squared[t] = results.rsquared_adj
        except Exception as e:
            # print(f"{e}")
            coefficients[t, :] = np.nan
            adjusted_r_squared[t] = np.nan

    # Calculate the arithmetic mean of adjusted_r_squared
    mean_R2 = np.nanmean(adjusted_r_squared)

    # Initialize array to hold the second stage results
    bhat = np.nan * np.ones(coefficients.shape[1])
    t_stat = np.nan * np.ones(coefficients.shape[1])

    # Second stage of FMB regression.
    for i in range(coefficients.shape[1]):
        y = coefficients[:, i]
        x = np.ones_like(y)

        if p['neweyWestLags'] > 0:
            # Newey-West estimator
            model = sm.OLS(y, x, missing='drop')
            results = model.fit(cov_type='HAC', cov_kwds={'maxlags': p['neweyWestLags']})
        else:
            # Simple OLS
            model = sm.OLS(y, x, missing='drop')
            results = model.fit()

        bhat[i] = results.params[0]
        t_stat[i] = results.tvalues[0]

    res = {
        'beta': coefficients,
        'bhat': bhat,
        't': t_stat,
        'Rbar2': adjusted_r_squared,
        'mean_R2': mean_R2,
    }

    if p['printResults'] == 1:
        prtFMBResults(p, res['bhat'], res['t'])

    return res


#
# from AssayingAnomalies import Config
# params = Config()
# params.set_up()
#
# y = pd.read_csv(params.crspFolder + '/ret.csv', index_col=0).astype(float).to_numpy()
# R = pd.read_csv(params.crspFolder + '/R.csv', index_col=0).astype(float)
# me = pd.read_csv(params.crspFolder + '/me.csv', index_col=0).astype(float)
# RHS_variables = [100*np.log(me), R]
# dates = pd.read_csv(params.crspFolder + '/dates.csv', index_col=0)
#
# results = run_Fama_MacBeth(y, RHS_variables, dates)
#
